# Copyright (C) 2015-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test the `mailman members` command."""

import unittest

from click.testing import CliRunner
from mailman.app.lifecycle import create_list
from mailman.commands.cli_members import members
from mailman.interfaces.member import MemberRole
from mailman.testing.helpers import subscribe
from mailman.testing.layers import ConfigLayer
from tempfile import NamedTemporaryFile


class TestCLIMembers(unittest.TestCase):
    layer = ConfigLayer

    def setUp(self):
        self._mlist = create_list('ant@example.com')
        self._command = CliRunner()

    def test_no_such_list(self):
        result = self._command.invoke(members, ('bee.example.com',))
        self.assertEqual(result.exit_code, 2)
        self.assertEqual(
            result.output,
            'Usage: members [OPTIONS] LISTSPEC\n'
            'Try \'members --help\' for help.\n\n'
            'Error: No such list: bee.example.com\n')

    def test_role_administrator(self):
        subscribe(self._mlist, 'Anne', role=MemberRole.owner)
        subscribe(self._mlist, 'Bart', role=MemberRole.moderator)
        subscribe(self._mlist, 'Cate', role=MemberRole.nonmember)
        subscribe(self._mlist, 'Dave', role=MemberRole.member)
        with NamedTemporaryFile('w', encoding='utf-8') as outfp:
            self._command.invoke(members, (
                '--role', 'administrator', '-o', outfp.name,
                'ant.example.com'))
            with open(outfp.name, 'r', encoding='utf-8') as infp:
                lines = infp.readlines()
        self.assertEqual(len(lines), 2)
        self.assertEqual(lines[0], 'Anne Person <aperson@example.com>\n')
        self.assertEqual(lines[1], 'Bart Person <bperson@example.com>\n')

    def test_role_any(self):
        subscribe(self._mlist, 'Anne', role=MemberRole.owner)
        subscribe(self._mlist, 'Bart', role=MemberRole.moderator)
        subscribe(self._mlist, 'Cate', role=MemberRole.nonmember)
        subscribe(self._mlist, 'Dave', role=MemberRole.member)
        with NamedTemporaryFile('w', encoding='utf-8') as outfp:
            self._command.invoke(members, (
                '--role', 'any', '-o', outfp.name, 'ant.example.com'))
            with open(outfp.name, 'r', encoding='utf-8') as infp:
                lines = infp.readlines()
        self.assertEqual(len(lines), 4)
        self.assertEqual(lines[0], 'Anne Person <aperson@example.com>\n')
        self.assertEqual(lines[1], 'Bart Person <bperson@example.com>\n')
        self.assertEqual(lines[2], 'Cate Person <cperson@example.com>\n')
        self.assertEqual(lines[3], 'Dave Person <dperson@example.com>\n')

    def test_role_moderator(self):
        subscribe(self._mlist, 'Anne', role=MemberRole.owner)
        subscribe(self._mlist, 'Bart', role=MemberRole.moderator)
        subscribe(self._mlist, 'Cate', role=MemberRole.nonmember)
        subscribe(self._mlist, 'Dave', role=MemberRole.member)
        with NamedTemporaryFile('w', encoding='utf-8') as outfp:
            self._command.invoke(members, (
                '--role', 'moderator', '-o', outfp.name, 'ant.example.com'))
            with open(outfp.name, 'r', encoding='utf-8') as infp:
                lines = infp.readlines()
        self.assertEqual(len(lines), 1)
        self.assertEqual(lines[0], 'Bart Person <bperson@example.com>\n')

    def test_role_nonmember(self):
        subscribe(self._mlist, 'Anne', role=MemberRole.owner)
        subscribe(self._mlist, 'Bart', role=MemberRole.moderator)
        subscribe(self._mlist, 'Cate', role=MemberRole.nonmember)
        subscribe(self._mlist, 'Dave', role=MemberRole.member)
        with NamedTemporaryFile('w', encoding='utf-8') as outfp:
            self._command.invoke(members, (
                '--role', 'nonmember', '-o', outfp.name, 'ant.example.com'))
            with open(outfp.name, 'r', encoding='utf-8') as infp:
                lines = infp.readlines()
        self.assertEqual(len(lines), 1)
        self.assertEqual(lines[0], 'Cate Person <cperson@example.com>\n')

    def test_already_subscribed_with_display_name(self):
        subscribe(self._mlist, 'Anne')
        with NamedTemporaryFile('w', buffering=1, encoding='utf-8') as infp:
            print('Anne Person <aperson@example.com>', file=infp)
            result = self._command.invoke(members, (
                '--add', infp.name, 'ant.example.com'))
        self.assertEqual(
            result.output,
            'Usage: members [OPTIONS] LISTSPEC\n'
            'Try \'members --help\' for help.\n\n'
            'Error: The --add option is removed. Use '
            '`mailman addmembers` instead.\n')

    def test_add_invalid_email(self):
        with NamedTemporaryFile('w', buffering=1, encoding='utf-8') as infp:
            print('foobar@', file=infp)
            result = self._command.invoke(members, (
                '--add', infp.name, 'ant.example.com'))
        self.assertEqual(
            result.output,
            'Usage: members [OPTIONS] LISTSPEC\n'
            'Try \'members --help\' for help.\n\n'
            'Error: The --add option is removed. Use '
            '`mailman addmembers` instead.\n')

    def test_not_subscribed_without_display_name(self):
        with NamedTemporaryFile('w', buffering=1, encoding='utf-8') as infp:
            print('aperson@example.com', file=infp)
            result = self._command.invoke(members, (
                '--delete', infp.name, 'ant.example.com'))
        self.assertEqual(
            result.output,
            'Usage: members [OPTIONS] LISTSPEC\n'
            'Try \'members --help\' for help.\n\n'
            'Error: The --delete option is removed. Use '
            '`mailman delmembers` instead.\n')

    def test_not_subscribed_with_display_name(self):
        with NamedTemporaryFile('w', buffering=1, encoding='utf-8') as infp:
            print('Anne Person <aperson@example.com>', file=infp)
            result = self._command.invoke(members, (
                '--delete', infp.name, 'ant.example.com'))
        self.assertEqual(
            result.output,
            'Usage: members [OPTIONS] LISTSPEC\n'
            'Try \'members --help\' for help.\n\n'
            'Error: The --delete option is removed. Use '
            '`mailman delmembers` instead.\n')

    def test_deletion_blank_lines(self):
        subscribe(self._mlist, 'Anne')
        subscribe(self._mlist, 'Bart')
        subscribe(self._mlist, 'Cate')
        with NamedTemporaryFile('w', buffering=1, encoding='utf-8') as infp:
            print('Anne Person <aperson@example.com>', file=infp)
            print('', file=infp)
            print('   ', file=infp)
            print('\t', file=infp)
            print('Bart Person <bperson@example.com>', file=infp)
            result = self._command.invoke(members, (
                '--delete', infp.name, 'ant.example.com'))
        self.assertEqual(
            result.output,
            'Usage: members [OPTIONS] LISTSPEC\n'
            'Try \'members --help\' for help.\n\n'
            'Error: The --delete option is removed. Use '
            '`mailman delmembers` instead.\n')
        with NamedTemporaryFile('w', encoding='utf-8') as outfp:
            self._command.invoke(members, (
                '-o', outfp.name, 'ant.example.com'))
            with open(outfp.name, 'r', encoding='utf-8') as infp:
                lines = infp.readlines()
        self.assertEqual(len(lines), 3)
        self.assertEqual(lines[0], 'Anne Person <aperson@example.com>\n')

    def test_deletion_commented_lines(self):
        subscribe(self._mlist, 'Anne')
        subscribe(self._mlist, 'Bart')
        with NamedTemporaryFile('w', buffering=1, encoding='utf-8') as infp:
            print('Anne Person <aperson@example.com>', file=infp)
            print('#Bart Person <bperson@example.com>', file=infp)
            result = self._command.invoke(members, (
                '--delete', infp.name, 'ant.example.com'))
        self.assertEqual(
            result.output,
            'Usage: members [OPTIONS] LISTSPEC\n'
            'Try \'members --help\' for help.\n\n'
            'Error: The --delete option is removed. Use '
            '`mailman delmembers` instead.\n')

        with NamedTemporaryFile('w', encoding='utf-8') as outfp:
            self._command.invoke(members, (
                '-o', outfp.name, 'ant.example.com'))
            with open(outfp.name, 'r', encoding='utf-8') as infp:
                lines = infp.readlines()
        self.assertEqual(len(lines), 2)
        self.assertEqual(lines[0], 'Anne Person <aperson@example.com>\n')

    def test_sync_invalid_email(self):
        with NamedTemporaryFile('w', buffering=1, encoding='utf-8') as infp:
            print('Dont Subscribe <not-a-valid-email>', file=infp)
            print('not-a-valid@email', file=infp)
            result = self._command.invoke(members, (
                '--sync', infp.name, 'ant.example.com'))
        self.assertEqual(
            result.output,
            'Usage: members [OPTIONS] LISTSPEC\n'
            'Try \'members --help\' for help.\n\n'
            'Error: The --sync option is removed. '
            'Use `mailman syncmembers` instead.\n')

    def test_sync_no_change(self):
        subscribe(self._mlist, 'Anne')
        subscribe(self._mlist, 'Bart')
        with NamedTemporaryFile('w', buffering=1, encoding='utf-8') as infp:
            print('Anne Person <aperson@example.com>', file=infp)
            result = self._command.invoke(members, (
                '--no-change', '--sync', infp.name, 'ant.example.com'))
        self.assertEqual(
            result.output,
            'Usage: members [OPTIONS] LISTSPEC\n'
            'Try \'members --help\' for help.\n\n'
            'Error: The --sync option is removed. '
            'Use `mailman syncmembers` instead.\n')

    def test_sync_empty_tuple(self):
        subscribe(self._mlist, 'Anne')
        subscribe(self._mlist, 'Bart')
        with NamedTemporaryFile('w', buffering=1, encoding='utf-8') as infp:
            print('Anne Person <aperson@example.com>', file=infp)
            print('\"\"', file=infp)
            result = self._command.invoke(members, (
                '--no-change', '--sync', infp.name, 'ant.example.com'))
        self.assertEqual(
            result.output,
            'Usage: members [OPTIONS] LISTSPEC\n'
            'Try \'members --help\' for help.\n\n'
            'Error: The --sync option is removed. '
            'Use `mailman syncmembers` instead.\n')

    def test_sync_commented_lines(self):
        subscribe(self._mlist, 'Anne')
        subscribe(self._mlist, 'Bart')
        with NamedTemporaryFile('w', buffering=1, encoding='utf-8') as infp:
            print('Anne Person <aperson@example.com>', file=infp)
            print('#Bart Person <bperson@example.com>', file=infp)
            result = self._command.invoke(members, (
                '--sync', infp.name, 'ant.example.com'))
        self.assertEqual(
            result.output,
            'Usage: members [OPTIONS] LISTSPEC\n'
            'Try \'members --help\' for help.\n\n'
            'Error: The --sync option is removed. '
            'Use `mailman syncmembers` instead.\n')

        with NamedTemporaryFile('w', encoding='utf-8') as outfp:
            self._command.invoke(members, (
                '-o', outfp.name, 'ant.example.com'))
            with open(outfp.name, 'r', encoding='utf-8') as infp:
                lines = infp.readlines()
        self.assertEqual(len(lines), 2)
        self.assertEqual(lines[0], 'Anne Person <aperson@example.com>\n')

    def test_sync_blank_lines(self):
        subscribe(self._mlist, 'Anne')
        subscribe(self._mlist, 'Bart')
        with NamedTemporaryFile('w', buffering=1, encoding='utf-8') as infp:
            print('Anne Person <aperson@example.com>', file=infp)
            print('', file=infp)
            print('', file=infp)
            result = self._command.invoke(members, (
                '--sync', infp.name, 'ant.example.com'))
        self.assertEqual(
            result.output,
            'Usage: members [OPTIONS] LISTSPEC\n'
            'Try \'members --help\' for help.\n\n'
            'Error: The --sync option is removed. '
            'Use `mailman syncmembers` instead.\n')

        with NamedTemporaryFile('w', encoding='utf-8') as outfp:
            self._command.invoke(members, (
                '-o', outfp.name, 'ant.example.com'))
            with open(outfp.name, 'r', encoding='utf-8') as infp:
                lines = infp.readlines()
        self.assertEqual(len(lines), 2)
        self.assertEqual(lines[0], 'Anne Person <aperson@example.com>\n')

    def test_sync_nothing_to_do(self):
        subscribe(self._mlist, 'Anne')
        subscribe(self._mlist, 'Bart')
        with NamedTemporaryFile('w', buffering=1, encoding='utf-8') as infp:
            print('Anne Person <aperson@example.com>', file=infp)
            print('Bart Person <bperson@example.com>', file=infp)
            result = self._command.invoke(members, (
                '--sync', infp.name, 'ant.example.com'))
        self.assertEqual(
            result.output,
            'Usage: members [OPTIONS] LISTSPEC\n'
            'Try \'members --help\' for help.\n\n'
            'Error: The --sync option is removed. '
            'Use `mailman syncmembers` instead.\n')

        with NamedTemporaryFile('w', encoding='utf-8') as outfp:
            self._command.invoke(members, (
                '-o', outfp.name, 'ant.example.com'))
            with open(outfp.name, 'r', encoding='utf-8') as infp:
                lines = infp.readlines()
        self.assertEqual(len(lines), 2)
        self.assertEqual(lines[0], 'Anne Person <aperson@example.com>\n')
        self.assertEqual(lines[1], 'Bart Person <bperson@example.com>\n')

    def test_sync_no_display_name(self):
        subscribe(self._mlist, 'Bart')
        subscribe(self._mlist, 'Cate', role=MemberRole.nonmember)
        with NamedTemporaryFile('w', buffering=1, encoding='utf-8') as infp:
            print('<aperson@example.com>', file=infp)
            result = self._command.invoke(members, (
                '--sync', infp.name, 'ant.example.com'))
        self.assertEqual(
            result.output,
            'Usage: members [OPTIONS] LISTSPEC\n'
            'Try \'members --help\' for help.\n\n'
            'Error: The --sync option is removed. '
            'Use `mailman syncmembers` instead.\n')

        with NamedTemporaryFile('w', encoding='utf-8') as outfp:
            self._command.invoke(members, (
                '-o', outfp.name, 'ant.example.com'))
            with open(outfp.name, 'r', encoding='utf-8') as infp:
                lines = infp.readlines()
        self.assertEqual(len(lines), 1)
        self.assertEqual(lines[0], 'Bart Person <bperson@example.com>\n')

    def test_sync_del_no_display_name(self):
        with NamedTemporaryFile('w', buffering=1, encoding='utf-8') as infp:
            print('bperson@example.com', file=infp)
            result = self._command.invoke(members, (
                '--add', infp.name, 'ant.example.com'))

        with NamedTemporaryFile('w', buffering=1, encoding='utf-8') as infp:
            print('<aperson@example.com>', file=infp)
            result = self._command.invoke(members, (
                '--sync', infp.name, 'ant.example.com'))
        self.assertEqual(
            result.output,
            'Usage: members [OPTIONS] LISTSPEC\n'
            'Try \'members --help\' for help.\n\n'
            'Error: The --sync option is removed. '
            'Use `mailman syncmembers` instead.\n')

        with NamedTemporaryFile('w', encoding='utf-8') as outfp:
            self._command.invoke(members, (
                '-o', outfp.name, 'ant.example.com'))
            with open(outfp.name, 'r', encoding='utf-8') as infp:
                lines = infp.readlines()
        self.assertEqual(len(lines), 1)
        self.assertEqual(lines[0], 'ant.example.com has no members\n')

    def test_email_only(self):
        subscribe(self._mlist, 'Anne')
        subscribe(self._mlist, 'Bart')
        result = self._command.invoke(members, (
            '--email-only', 'ant.example.com'))
        self.assertEqual(
            result.output, 'aperson@example.com\nbperson@example.com\n')

    def test_count_only(self):
        subscribe(self._mlist, 'Anne')
        subscribe(self._mlist, 'Bart')
        result = self._command.invoke(members, (
            '--count-only', 'ant.example.com'))
        self.assertEqual(
            result.output, '2\n')

    def test_incompatible_email_only_count_only(self):
        result = self._command.invoke(members, (
            '--email-only', '--count-only', 'ant.example.com'))
        self.assertEqual(
            result.output,
            'Usage: members [OPTIONS] LISTSPEC\n'
            'Try \'members --help\' for help.\n\n'
            'Error: The --email_only and --count_only options are '
            'mutually exclusive.\n')

    def test_incompatible_role_any_regular(self):
        result = self._command.invoke(members, (
            '--role', 'any', '--regular', 'ant.example.com'))
        self.assertEqual(
            result.output,
            'Usage: members [OPTIONS] LISTSPEC\n'
            'Try \'members --help\' for help.\n\n'
            'Error: The --regular, --digest and --nomail options are '
            'incompatible with role=any.\n')

    def test_incompatible_role_any_digest(self):
        result = self._command.invoke(members, (
            '--role', 'any', '--digest', 'any', 'ant.example.com'))
        self.assertEqual(
            result.output,
            'Usage: members [OPTIONS] LISTSPEC\n'
            'Try \'members --help\' for help.\n\n'
            'Error: The --regular, --digest and --nomail options are '
            'incompatible with role=any.\n')

    def test_incompatible_role_any_nomail(self):
        result = self._command.invoke(members, (
            '--role', 'any', '--nomail', 'any', 'ant.example.com'))
        self.assertEqual(
            result.output,
            'Usage: members [OPTIONS] LISTSPEC\n'
            'Try \'members --help\' for help.\n\n'
            'Error: The --regular, --digest and --nomail options are '
            'incompatible with role=any.\n')

    def test_non_ascii_display_name(self):
        subscribe(self._mlist, 'Bögüs', role=MemberRole.member)
        with NamedTemporaryFile('w', encoding='utf-8') as outfp:
            self._command.invoke(members, (
                '-o', outfp.name, 'ant.example.com'))
            with open(outfp.name, 'r', encoding='utf-8') as infp:
                lines = infp.readlines()
        self.assertEqual(len(lines), 1)
        self.assertEqual(lines[0], 'Bögüs Person <bperson@example.com>\n')



# Copyright (C) 2015-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test the discard chain."""

import unittest

from mailman.app.lifecycle import create_list
from mailman.core.chains import process as process_chain
from mailman.testing.helpers import (
    LogFileMark,
    specialized_message_from_string as mfs,
)
from mailman.testing.layers import ConfigLayer


class TestDiscard(unittest.TestCase):
    """Test the discard chain."""

    layer = ConfigLayer

    def setUp(self):
        self._mlist = create_list('test@example.com')
        self._msg = mfs("""\
From: anne@example.com
To: test@example.com
Subject: Ignore
Message-Id: <mid@example.com>

""")

    def test_discard_reasons(self):
        # The log message must contain the moderation reasons.
        msgdata = dict(moderation_reasons=['TEST-REASON-1', 'TEST-REASON-2'])
        log_file = LogFileMark('mailman.vette')
        process_chain(self._mlist, self._msg, msgdata, start_chain='discard')
        log_entry = log_file.read()
        self.assertIn('DISCARD: <mid@example.com>', log_entry)
        self.assertIn('TEST-REASON-1', log_entry)
        self.assertIn('TEST-REASON-2', log_entry)

    def test_discard_no_reasons(self):
        # The log message contains n/a if no moderation reasons.
        msgdata = {}
        log_file = LogFileMark('mailman.vette')
        process_chain(self._mlist, self._msg, msgdata, start_chain='discard')
        log_entry = log_file.read()
        self.assertIn('DISCARD: <mid@example.com>', log_entry)
        self.assertIn('[n/a]', log_entry)



# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test posting to a mailing list's -owner address."""

# XXX 2012-03-23 BAW: This is not necessarily the best place for this test.
# We really need a better place to collect these sort of end-to-end posting
# tests.  They're not exactly integration tests, but they do touch lots of
# parts of the system.

import unittest

from mailman.app.lifecycle import create_list
from mailman.config import config
from mailman.database.transaction import transaction
from mailman.interfaces.member import MemberRole
from mailman.interfaces.usermanager import IUserManager
from mailman.runners.incoming import IncomingRunner
from mailman.runners.outgoing import OutgoingRunner
from mailman.runners.pipeline import PipelineRunner
from mailman.testing.helpers import (
    get_lmtp_client,
    make_testable_runner,
    TestableMaster,
)
from mailman.testing.layers import SMTPLayer
from operator import itemgetter
from zope.component import getUtility


class TestEmailToOwner(unittest.TestCase):
    """Test emailing a mailing list's -owner address."""

    layer = SMTPLayer

    def setUp(self):
        self._mlist = create_list('test@example.com')
        # Add some owners, moderators, and members
        manager = getUtility(IUserManager)
        with transaction():
            anne = manager.create_address('anne@example.com')
            bart = manager.create_address('bart@example.com')
            cris = manager.create_address('cris@example.com')
            dave = manager.create_address('dave@example.com')
            self._mlist.subscribe(anne, MemberRole.member)
            self._mlist.subscribe(anne, MemberRole.owner)
            self._mlist.subscribe(bart, MemberRole.moderator)
            self._mlist.subscribe(bart, MemberRole.owner)
            self._mlist.subscribe(cris, MemberRole.moderator)
            self._mlist.subscribe(dave, MemberRole.member)
        self._inq = make_testable_runner(IncomingRunner, 'in')
        self._pipelineq = make_testable_runner(PipelineRunner, 'pipeline')
        self._outq = make_testable_runner(OutgoingRunner, 'out')

    def test_owners_get_email(self):
        # XXX 2012-03-23 BAW: We can't use a layer here because we need both
        # the SMTPLayer and LMTPLayer and these are incompatible.  There's no
        # way to make zope.test* happy without causing errors or worse.  Live
        # with this hack until we can rip all that layer crap out and use
        # something like testresources.
        def wait():
            get_lmtp_client(quiet=True)
        lmtpd = TestableMaster(wait)
        lmtpd.start('lmtp')
        # Post a message to the list's -owner address, and all the owners will
        # get a copy of the message.
        lmtp = get_lmtp_client(quiet=True)
        lmtp.lhlo('remote.example.org')
        lmtp.sendmail('zuzu@example.org', ['test-owner@example.com'], b"""\
From: Zuzu Person <zuzu@example.org>
To: test-owner@example.com
Message-ID: <ant>

Can you help me?
""")
        lmtpd.stop()
        # There should now be one message sitting in the incoming queue.
        # Check that, then process it.  Don't use get_queue_messages() since
        # that will empty the queue.
        self.assertEqual(len(config.switchboards['in'].files), 1)
        self._inq.run()
        # There should now be one message sitting in the pipeline queue.
        # Process that one too.
        self.assertEqual(len(config.switchboards['pipeline'].files), 1)
        self._pipelineq.run()
        # The message has made its way to the outgoing queue.  Again, check
        # and process that one.
        self.assertEqual(len(config.switchboards['out'].files), 1)
        self._outq.run()
        # The SMTP server has now received three messages, one for each of the
        # owners and moderators.  Of course, Bart is both an owner and a
        # moderator, so he'll get only one copy of the message.  Dave does not
        # get a copy of the message.
        messages = sorted(SMTPLayer.smtpd.messages, key=itemgetter('x-rcptto'))
        self.assertEqual(len(messages), 3)
        self.assertEqual(messages[0]['x-rcptto'], 'anne@example.com')
        self.assertEqual(messages[1]['x-rcptto'], 'bart@example.com')
        self.assertEqual(messages[2]['x-rcptto'], 'cris@example.com')
        # And yet, all three messages are addressed to the -owner address.
        for message in messages:
            self.assertEqual(message['to'], 'test-owner@example.com')
        # All three messages will have two X-MailFrom headers.  One is added
        # by the LMTP server accepting Zuzu's original message, and will
        # contain her posting address, i.e. zuzu@example.com.  The second one
        # is added by the aiosmtpd server that accepts Mailman's VERP'd
        # message to the individual recipient.  By verifying both, we prove
        # that Zuzu sent the original message, and that Mailman is VERP'ing
        # the copy to all the owners.
        self.assertEqual(
            messages[0].get_all('x-mailfrom'),
            ['zuzu@example.org', 'test-bounces+anne=example.com@example.com'])
        self.assertEqual(
            messages[1].get_all('x-mailfrom'),
            ['zuzu@example.org', 'test-bounces+bart=example.com@example.com'])
        self.assertEqual(
            messages[2].get_all('x-mailfrom'),
            ['zuzu@example.org', 'test-bounces+cris=example.com@example.com'])



# Copyright (C) 2009-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Fake MTA for testing purposes."""
from __future__ import generator_stop

import smtplib

from aiosmtpd.controller import Controller
from aiosmtpd.handlers import Message as MessageHandler
from aiosmtpd.smtp import SMTP
from mailman.interfaces.mta import IMailTransportAgentLifecycle
from public import public
from queue import Empty, Queue
from zope.interface import implementer


@public
@implementer(IMailTransportAgentLifecycle)
class FakeMTA:
    """Fake MTA for testing purposes."""

    def create(self, mlist):
        pass

    def delete(self, mlist):
        pass

    def regenerate(self, directory=None):
        pass


class ConnectionCountingHandler(MessageHandler):
    def __init__(self, msg_queue):
        super().__init__()
        self._msg_queue = msg_queue
        self.connection_count = 0

    def handle_message(self, message):
        self._msg_queue.put(message)

    async def handle_EHLO(self, server, session, envelope, hostname):
        session.host_name = hostname
        await server.push('250-AUTH PLAIN')
        return '250 HELP'

    async def handle_RSET(self, server, session, envelope):
        self.connection_count = 0
        return '250 OK'


class ConnectionCountingSMTP(SMTP):
    def __init__(self, handler, oob_queue, err_queue, *args, **kws):
        super().__init__(handler, *args, **kws)
        self._auth_response = None
        self._waiting_for_auth_response = False
        self._oob_queue = oob_queue
        self._err_queue = err_queue
        self._last_error = None

    def connection_made(self, transport):
        super().connection_made(transport)
        # We can't keep the connection count on self here because the
        # controller (via the factory() method) will create a new instance of
        # this class for every connection.  The handler instance is always the
        # same though, so it's fine to stash this value away there.
        self.event_handler.connection_count += 1

    async def smtp_AUTH(self, arg):
        """Record that the AUTH occurred."""
        args = arg.split()
        if args[0].lower() == 'plain':
            if len(args) == 2:
                response = args[1]
                # The second argument is the AUTH PLAIN <initial-response>
                # which must be equal to the base 64 equivalent of the
                # expected login string "testuser:testpass".
                if response == 'AHRlc3R1c2VyAHRlc3RwYXNz':
                    await self.push('235 Ok')
                    self._oob_queue.put(response)
                else:
                    await self.push('571 Bad authentication')
            else:
                assert len(args) == 1, args
                # Send a challenge and set us up to wait for the response.
                await self.push('334 ')
                self._waiting_for_auth_response = True
        else:
            await self.push('571 Bad authentication')

    async def smtp_STAT(self, arg):
        """Cause the server to send statistics to its controller."""
        # Do not count the connection caused by the STAT connect.
        self.event_handler.connection_count -= 1
        self._oob_queue.put(self.event_handler.connection_count)
        await self.push('250 Ok')

    def _next_error(self, command):
        """Return the next error for the SMTP command, if there is one.

        :param command: The SMTP command for which an error might be
            expected.  If the next error matches the given command, the
            expected error code is returned.
        :type command: string, lower-cased
        :return: An SMTP error code
        :rtype: integer
        """
        # If the last error we pulled from the queue didn't match, then we're
        # caching it, and it might match this expected error.  If there is no
        # last error in the cache, get one from the queue now.
        if self._last_error is None:
            try:
                self._last_error = self._err_queue.get_nowait()
            except Empty:
                # No error is expected
                return None
        if self._last_error[0] == command:
            code = self._last_error[1]
            self._last_error = None
            return code
        return None

    async def smtp_RCPT(self, arg):
        """For testing, sometimes cause a non-25x response."""
        code = self._next_error('rcpt')
        if code is None:
            # Everything's cool.
            await super().smtp_RCPT(arg)
        else:
            # The test suite wants this to fail.  The message corresponds to
            # the exception we expect smtplib.SMTP to raise.
            await self.push('%d Error: SMTPRecipientsRefused' % code)

    async def smtp_MAIL(self, arg):
        """For testing, sometimes cause a non-25x response."""
        code = self._next_error('mail')
        if code is None:
            # Everything's cool.
            await super().smtp_MAIL(arg)
        else:
            # The test suite wants this to fail.  The message corresponds to
            # the exception we expect smtplib.SMTP to raise.
            await self.push('%d Error: SMTPResponseException' % code)


class ConnectionCountingController(Controller):
    """Count the number of SMTP connections opened."""

    def __init__(self, host, port, ssl_context=None):
        self._msg_queue = Queue()
        self._oob_queue = Queue()
        self.err_queue = Queue()

        handler = ConnectionCountingHandler(self._msg_queue)
        super().__init__(handler, hostname=host, port=port,
                         ssl_context=ssl_context)

    def factory(self):
        return ConnectionCountingSMTP(
            self.handler, self._oob_queue, self.err_queue)

    def start(self):
        super().start()
        # Reset the connection statistics, since the base class's start()
        # method causes a connection to occur.
        self.reset()

    def _connect(self):
        client = smtplib.SMTP(self.hostname, self.port)
        return client

    def get_connection_count(self):
        """Retrieve the number of connections.

        :return: The number of connections to the server that have been made.
        :rtype: integer
        """
        client = self._connect()
        client.docmd('STAT')
        # An Empty exception will occur if the data isn't available in 10
        # seconds.  Let that propagate.
        return self._oob_queue.get(block=True, timeout=10)

    def get_authentication_credentials(self):
        """Retrieve the last authentication credentials."""
        return self._oob_queue.get(block=True, timeout=10)

    def __iter__(self):
        while True:
            try:
                yield self._msg_queue.get_nowait()
            except Empty:
                return

    @property
    def messages(self):
        """Return all the messages received by the SMTP server."""
        yield from self

    def clear(self):
        """Clear all the messages from the queue."""
        list(self)

    def reset(self):
        client = self._connect()
        client.docmd('RSET')


class ConnectionCountingSSLController(ConnectionCountingController):
    """Count the number of SMTPS connections opened."""

    def __init__(self, host, port, client_context=None, server_context=None):
        super().__init__(host, port, ssl_context=server_context)
        self._client_context = client_context

    def _connect(self):
        client = smtplib.SMTP_SSL(self.hostname, self.port,
                                  context=self._client_context)
        return client


class ConnectionCountingSTARTTLSController(ConnectionCountingController):
    """Count the number of SMTP connections with STARTTLS opened."""

    def __init__(self, host, port, client_context=None, server_context=None):
        super().__init__(host, port)
        self._client_context = client_context
        self._server_context = server_context

    def factory(self):
        return ConnectionCountingSMTP(
            self.handler, self._oob_queue,
            self.err_queue, tls_context=self._server_context,
            require_starttls=True)

    def _connect(self):
        client = smtplib.SMTP(self.hostname, self.port)
        client.starttls(context=self._client_context)
        return client

    def get_connection_count(self):
        count = super().get_connection_count()
        # This is a hack, since when using STARTTLS for every connection the
        # SMTP.connection_made(transport) method is called twice.
        # The -1 is there for the last pair of connections made when checking
        # the connection count, since one of them is already subtracted in
        # ConnectionCountingSMTP.handle_STAT.
        # The //2 is there for the rest of the connections which are counted
        # twice.
        return (count - 1) // 2



# Copyright (C) 2016-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test address bans."""

import unittest

from mailman.app.lifecycle import create_list
from mailman.database.transaction import transaction
from mailman.interfaces.bans import IBanManager
from mailman.testing.helpers import call_api
from mailman.testing.layers import RESTLayer
from urllib.error import HTTPError


class TestBans(unittest.TestCase):
    layer = RESTLayer

    def setUp(self):
        with transaction():
            self._mlist = create_list('ant@example.com')

    def test_get_missing_banned_address(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/lists/ant.example.com'
                     '/bans/notbanned@example.com')
        self.assertEqual(cm.exception.code, 404)
        self.assertEqual(cm.exception.reason,
                         'Email is not banned: notbanned@example.com')

    def test_delete_missing_banned_address(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/lists/ant.example.com'
                     '/bans/notbanned@example.com',
                     method='DELETE')
        self.assertEqual(cm.exception.code, 404)
        self.assertEqual(cm.exception.reason,
                         'Email is not banned: notbanned@example.com')

    def test_not_found_after_unbanning(self):
        manager = IBanManager(self._mlist)
        with transaction():
            manager.ban('banned@example.com')
        url = ('http://localhost:9001/3.0/lists/ant.example.com'
               '/bans/banned@example.com')
        json, response = call_api(url)
        self.assertEqual(json['email'], 'banned@example.com')
        json, response = call_api(url, method='DELETE')
        self.assertEqual(response.status_code, 204)
        with self.assertRaises(HTTPError) as cm:
            call_api(url)
        self.assertEqual(cm.exception.code, 404)
        self.assertEqual(cm.exception.reason,
                         'Email is not banned: banned@example.com')

    def test_not_found_after_unbanning_global(self):
        manager = IBanManager(None)
        with transaction():
            manager.ban('banned@example.com')
        url = ('http://localhost:9001/3.0/bans/banned@example.com')
        json, response = call_api(url)
        self.assertEqual(json['email'], 'banned@example.com')
        json, response = call_api(url, method='DELETE')
        self.assertEqual(response.status_code, 204)
        with self.assertRaises(HTTPError) as cm:
            call_api(url)
        self.assertEqual(cm.exception.code, 404)
        self.assertEqual(cm.exception.reason,
                         'Email is not banned: banned@example.com')

    def test_ban_missing_mailing_list(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/lists/bee.example.com/bans')
        self.assertEqual(cm.exception.code, 404)

    def test_invalid_ban_email_address(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/lists/ant.example.com/bans',
                     dict(email='badaddress@example'),
                     method="POST")
        self.assertEqual(cm.exception.code, 400)
        # Workaround for coming API exception message change
        self.assertEqual(
            cm.exception.reason,
            'Invalid Parameter "email": Expected a valid email '
            'address or regular expression, got badaddress@example.')

    def test_invalid_ban_regex(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/lists/ant.example.com/bans',
                     dict(email='^[^@]+@(?!('),
                     method="POST")
        self.assertEqual(cm.exception.code, 400)
        # Workaround for coming API exception message change
        self.assertEqual(
            cm.exception.reason,
            'Invalid Parameter "email": Expected a valid email address'
            ' or regular expression, got ^[^@]+@(?!(.')

    def test_ban_regex_and_encoded_self_link(self):
        call_api('http://localhost:9001/3.0/lists/ant.example.com/bans',
                 dict(email='^[^@]+'),
                 method="POST")
        json, content = call_api(
            'http://localhost:9001/3.0/lists/ant.example.com'
            '/bans/^[^@]+')
        self.assertEqual(json['self_link'],
                         'http://localhost:9001/3.0/lists/ant.example.com'
                         '/bans/%5E%5B%5E%40%5D%2B')



# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test mailing list joins."""

import re
import unittest

from email.iterators import body_line_iterator
from mailman.app.lifecycle import create_list
from mailman.config import config
from mailman.interfaces.bans import IBanManager
from mailman.interfaces.member import DeliveryMode
from mailman.interfaces.subscriptions import (
    ISubscriptionManager,
    ISubscriptionService,
    TokenOwner,
)
from mailman.interfaces.usermanager import IUserManager
from mailman.runners.command import CommandRunner
from mailman.testing.helpers import (
    get_queue_messages,
    make_testable_runner,
    specialized_message_from_string as mfs,
)
from mailman.testing.layers import ConfigLayer
from zope.component import getUtility


class TestJoin(unittest.TestCase):
    """Test mailing list joins."""

    layer = ConfigLayer

    def setUp(self):
        self._mlist = create_list('test@example.com')
        self._mlist.send_welcome_message = False
        self._commandq = config.switchboards['command']
        self._runner = make_testable_runner(CommandRunner, 'command')

    def test_double_confirmation(self):
        # A join request comes in using both the -join address and the word
        # 'subscribe' in the first line of the body.  This should produce just
        # one subscription request.
        msg = mfs("""\
From: anne@example.org
To: test-join@example.com

subscribe
""")
        # Adding the subaddress to the metadata dictionary mimics what happens
        # when the above email message is first processed by the lmtp runner.
        # For convenience, we skip that step in this test.
        self._commandq.enqueue(msg, dict(listid='test.example.com',
                                         subaddress='join'))
        self._runner.run()
        # There will be one messages in the queue. This is the confirmation
        # message of her join request.
        items = get_queue_messages('virgin', sort_on='subject',
                                   expected_count=1)

        self.assertTrue(str(items[0].msg['subject']).startswith('Your conf'))

    def test_join_when_already_a_member(self):
        anne = getUtility(IUserManager).create_user('anne@example.org')
        self._mlist.subscribe(list(anne.addresses)[0])
        # When someone tries to join by email and they are already a member,
        # ignore the request.
        msg = mfs("""\
From: anne@example.org
To: test-join@example.com
Subject: join

""")
        self._commandq.enqueue(msg, dict(listid='test.example.com'))
        self._runner.run()
        # There will be one message in the queue - a reply to Anne notifying
        # her of the status of her command email.  Because Anne is already
        # subscribed to the list, she gets and needs no confirmation.
        items = get_queue_messages('virgin', expected_count=1)
        self.assertEqual(items[0].msg['subject'],
                         'The results of your email commands')
        # Search the contents of the results message.  There should be just
        # one 'Confirmation email' line.
        confirmation_lines = []
        in_results = False
        for line in body_line_iterator(items[0].msg):
            line = line.strip()
            if in_results:
                if line.startswith('- Done'):
                    break
                if len(line) > 0:
                    confirmation_lines.append(line)
            if line.strip() == '- Results:':
                in_results = True
        # There should be exactly one confirmation line.
        self.assertEqual(len(confirmation_lines), 1)
        # And the confirmation line should name Anne's email address.
        self.assertIn('anne@example.org', confirmation_lines[0])

    def test_join_banned_address_produces_error(self):
        msg = mfs("""\
From: anne@example.org
To: test-join@example.com
""")
        IBanManager(self._mlist).ban('anne@example.org')
        get_queue_messages('virgin')
        self._commandq.enqueue(msg, dict(listid='test.example.com',
                                         subaddress='join'))
        self._runner.run()
        # This should send out an error email.
        get_queue_messages('virgin', expected_count=1)

    def test_join_subscription_pending_produces_error(self):
        msg = mfs("""\
From: anne@example.org
To: test-join@example.com
""")
        # Pend a subscription.
        self._commandq.enqueue(msg, dict(listid='test.example.com',
                                         subaddress='join'))
        self._runner.run()
        # Clear the virgin queue.
        get_queue_messages('virgin')
        # And try another.
        self._commandq.enqueue(msg, dict(listid='test.example.com',
                                         subaddress='join'))
        self._runner.run()
        # This should send out an error email.
        get_queue_messages('virgin', expected_count=1)


class TestJoinWithDigests(unittest.TestCase):
    """Test mailing list joins with the digests=<no|mime|plain> argument."""

    layer = ConfigLayer

    def setUp(self):
        self._mlist = create_list('test@example.com')
        self._commandq = config.switchboards['command']
        self._runner = make_testable_runner(CommandRunner, 'command')

    def _confirm(self):
        # There will be two messages in the queue - the confirmation messages,
        # and a reply to Anne notifying her of the status of her command
        # email.  We need to dig the confirmation token out of the Subject
        # header of the latter so that we can confirm the subscription.
        items = get_queue_messages('virgin', sort_on='subject',
                                   expected_count=2)
        subject_words = str(items[1].msg['subject']).split()
        self.assertEqual(subject_words[0], 'Your')
        token = re.sub(r'^.*\+([^+@]*)@.*$', r'\1', str(items[1].msg['from']))
        token, token_owner, rmember = ISubscriptionManager(
            self._mlist).confirm(token)
        self.assertIsNone(token)
        self.assertEqual(token_owner, TokenOwner.no_one)
        # Now, make sure that Anne is a member of the list and is receiving
        # digest deliveries.
        members = getUtility(ISubscriptionService).find_members(
            'anne@example.org')
        self.assertEqual(len(members), 1)
        self.assertEqual(rmember, members[0])
        return rmember

    def test_join_in_subject_case_insensitive(self):
        # Test that join commands are case insensitive.
        msg = mfs("""\
From: anne@example.org
To: test-requests@example.com
Subject: Join

""")
        self._commandq.enqueue(msg, dict(listid='test.example.com'))
        self._runner.run()
        anne = self._confirm()
        self.assertEqual(anne.address.email, 'anne@example.org')
        self.assertEqual(anne.delivery_mode, DeliveryMode.regular)

    def test_join_case_insensitive(self):
        # Test that join commands are case insensitive.
        msg = mfs("""\
From: anne@example.org
To: test-requests@example.com

Join
""")
        self._commandq.enqueue(msg, dict(listid='test.example.com'))
        self._runner.run()
        anne = self._confirm()
        self.assertEqual(anne.address.email, 'anne@example.org')
        self.assertEqual(anne.delivery_mode, DeliveryMode.regular)

    def test_join_with_implicit_no_digests(self):
        # Test the digest=mime argument to the join command.
        msg = mfs("""\
From: anne@example.org
To: test-request@example.com

join
""")
        self._commandq.enqueue(msg, dict(listid='test.example.com'))
        self._runner.run()
        anne = self._confirm()
        self.assertEqual(anne.address.email, 'anne@example.org')
        self.assertEqual(anne.delivery_mode, DeliveryMode.regular)

    def test_join_with_explicit_no_digests(self):
        # Test the digest=mime argument to the join command.
        msg = mfs("""\
From: anne@example.org
To: test-request@example.com

join digest=no
""")
        self._commandq.enqueue(msg, dict(listid='test.example.com'))
        self._runner.run()
        anne = self._confirm()
        self.assertEqual(anne.address.email, 'anne@example.org')
        self.assertEqual(anne.delivery_mode, DeliveryMode.regular)

    # LP: #1444184 - digest=mime is not currently supported.
    def test_join_with_mime_digests(self):
        # Test the digest=mime argument to the join command.
        msg = mfs("""\
From: anne@example.org
To: test-request@example.com

join digest=mime
""")
        self._commandq.enqueue(msg, dict(listid='test.example.com'))
        self._runner.run()
        anne = self._confirm()
        self.assertEqual(anne.address.email, 'anne@example.org')
        self.assertEqual(anne.delivery_mode, DeliveryMode.mime_digests)

    # LP: #1444184 - digest=mime is not currently supported.
    def test_join_with_plain_digests(self):
        # Test the digest=mime argument to the join command.
        msg = mfs("""\
From: anne@example.org
To: test-request@example.com

join digest=plain
""")
        self._commandq.enqueue(msg, dict(listid='test.example.com'))
        self._runner.run()
        anne = self._confirm()
        self.assertEqual(anne.address.email, 'anne@example.org')
        self.assertEqual(anne.delivery_mode, DeliveryMode.plaintext_digests)



# Copyright (C) 2007-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Sending notifications."""

import logging

from email.mime.message import MIMEMessage
from email.mime.text import MIMEText
from email.utils import formataddr
from lazr.config import as_boolean
from mailman.config import config
from mailman.core.i18n import _
from mailman.email.message import OwnerNotification, UserNotification
from mailman.interfaces.member import DeliveryMode
from mailman.interfaces.messages import IMessageStore
from mailman.interfaces.template import ITemplateLoader
from mailman.utilities.string import expand, wrap
from public import public
from zope.component import getUtility


log = logging.getLogger('mailman.error')


def _get_dsn(message_id):
    # Get the DSN from the message store and delete it from the message store.
    messagestore = getUtility(IMessageStore)
    dsn = messagestore.get_message_by_id(message_id)
    if dsn:
        messagestore.delete_message(message_id)
    return dsn


def _make_multipart(msg):
    """Turn an OwnerNotification into a multipart/mixed with a text/plain
    payload. This is all a Kludge due to messages being email.Message objects
    and not email.EmailMessage objects. If they were the latter we could just
    use the message's add_content method in a straight forward way.
    """
    text = MIMEText(msg.get_payload(decode=True).decode(
                    msg.get_content_charset('utf-8')))
    del msg['content-type']
    del msg['content-transfer-encoding']
    msg['Content-Type'] = 'multipart/mixed'
    msg.set_payload(None)
    msg.attach(text)
    return msg


@public
def send_welcome_message(mlist, member, language, text=''):
    """Send a welcome message to a subscriber.

    Prepending to the standard welcome message template is the mailing list's
    welcome message, if there is one.

    :param mlist: The mailing list.
    :type mlist: IMailingList
    :param member: The member to send the welcome message to.
    :param address: IMember
    :param language: The language of the response.
    :type language: ILanguage
    """
    welcome_message = wrap(getUtility(ITemplateLoader).get(
        'list:user:notice:welcome', mlist, language=language.code))
    display_name = member.display_name
    # Get the text from the template.
    text = expand(welcome_message, mlist, dict(
        user_name=display_name,
        user_email=member.address.email,
        # For backward compatibility.
        user_address=member.address.email,
        fqdn_listname=mlist.fqdn_listname,
        list_name=mlist.display_name,
        list_requests=mlist.request_address,
        ))
    with _.using(language.code):
        digmode = (''                                   # noqa: F841
                   if member.delivery_mode is DeliveryMode.regular
                   else _(' (Digest mode)'))
        msg = UserNotification(
            formataddr((display_name, member.address.email)),
            mlist.request_address,
            _('Welcome to the "${mlist.display_name}" mailing list${digmode}'),
            text, language)
    msg['X-No-Archive'] = 'yes'
    msg.send(mlist, verp=as_boolean(config.mta.verp_personalized_deliveries))


@public
def send_goodbye_message(mlist, address, language):
    """Send a goodbye message to a subscriber.

    Prepending to the standard goodbye message template is the mailing list's
    goodbye message, if there is one.

    :param mlist: the mailing list
    :type mlist: IMailingList
    :param address: The address to respond to
    :type address: string
    :param language: the language of the response
    :type language: ILanguage
    """
    goodbye_message = wrap(expand(getUtility(ITemplateLoader).get(
        'list:user:notice:goodbye', mlist, language=language.code),
        mlist, dict(
        user_email=address,
        # For backward compatibility.
        user_address=address,
        fqdn_listname=mlist.fqdn_listname,
        list_name=mlist.display_name,
        list_requests=mlist.request_address,
        )))
    with _.using(language.code):
        msg = UserNotification(
            address, mlist.bounces_address,
            _('You have been unsubscribed from the ${mlist.display_name} '
              'mailing list'),
            goodbye_message, language)
    msg.send(mlist, verp=as_boolean(config.mta.verp_personalized_deliveries))


@public
def send_admin_subscription_notice(mlist, address, display_name):
    """Send the list administrators a subscription notice.

    :param mlist: The mailing list.
    :type mlist: IMailingList
    :param address: The address being subscribed.
    :type address: string
    :param display_name: The name of the subscriber.
    :type display_name: string
    """
    with _.using(mlist.preferred_language.code):
        subject = _('${mlist.display_name} subscription notification')
    text = expand(
        getUtility(ITemplateLoader).get('list:admin:notice:subscribe', mlist),
        mlist, dict(
            member=formataddr((display_name, address)),
            ))
    msg = OwnerNotification(mlist, subject, text, roster=mlist.administrators)
    msg.send(mlist)


@public
def send_admin_disable_notice(mlist, event, display_name):
    """Send the list administrators a membership disabled by-bounce notice.

    :param mlist: The mailing list
    :type mlist: IMailingList
    :param event: The BounceEvent that triggered this notice.
    :type event: A mailman.model.bounce.BounceEvent instance.
    :param display_name: The name of the subscriber
    :type display_name: string
    """
    member = formataddr((display_name, event.email))
    data = {'member': member}
    with _.using(mlist.preferred_language.code):
        subject = _(
            '${member}\'s subscription disabled on ${mlist.display_name}')
    text = expand(
        getUtility(ITemplateLoader).get('list:admin:notice:disable', mlist),
        mlist, data)
    msg = OwnerNotification(mlist, subject, text, roster=mlist.administrators)
    dsn = _get_dsn(event.message_id)
    if dsn:
        msg = _make_multipart(msg)
        att = MIMEMessage(dsn)
        msg.attach(att)
    msg.send(mlist)


@public
def send_admin_increment_notice(mlist, event, display_name):
    """Send the list administrators a bounce score incremented notice.

    :param mlist: The mailing list
    :type mlist: IMailingList
    :param event: The BounceEvent that triggered this notice.
    :type event: A mailman.model.bounce.BounceEvent instance.
    :param display_name: The name of the subscriber
    :type display_name: string
    """
    member = formataddr((display_name, event.email))
    data = {'member': member}
    with _.using(mlist.preferred_language.code):
        subject = _(
            '${member}\'s bounce score incremented on ${mlist.display_name}')
    text = expand(
        getUtility(ITemplateLoader).get('list:admin:notice:increment', mlist),
        mlist, data)
    msg = OwnerNotification(mlist, subject, text, roster=mlist.administrators)
    dsn = _get_dsn(event.message_id)
    if dsn:
        msg = _make_multipart(msg)
        att = MIMEMessage(dsn)
        msg.attach(att)
    msg.send(mlist)


def send_admin_removal_notice(mlist, address, display_name):
    """Send the list administrators a membership removed due to bounce notice.

    :param mlist: The mailing list.
    :type mlist: IMailingList
    :param address: The address of the member
    :type address: string
    :param display_name: The name of the subscriber
    :type display_name: string
    """
    member = formataddr((display_name, address))
    data = {'member': member, 'mlist': mlist.display_name}
    with _.using(mlist.preferred_language.code):
        subject = _('${member} unsubscribed from ${mlist.display_name} '
                    'mailing list due to bounces')
    text = expand(
        getUtility(ITemplateLoader).get('list:admin:notice:removal', mlist),
        mlist, data)
    msg = OwnerNotification(mlist, subject, text, roster=mlist.administrators)
    msg.send(mlist)


@public
def send_user_disable_warning(mlist, address, language):
    """Sends a warning mail to the user reminding the person to
    reenable its DeliveryStatus.

    :param mlist: The mailing list
    :type mlist: IMailingList
    :param address: The address of the member
    :type address: string.
    :param language: member's preferred language
    :type language: ILanguage
    """
    warning_message = wrap(getUtility(ITemplateLoader).get(
        'list:user:notice:warning', mlist, language=language.code))
    warning_message_text = expand(
        warning_message, mlist, dict(sender_email=address))
    msg = UserNotification(
        address, mlist.bounces_address,
        _('Your subscription for ${mlist.display_name} mailing list'
          ' has been disabled'),
        warning_message_text, language)
    msg.send(mlist, verp=as_boolean(config.mta.verp_personalized_deliveries))



# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test the archive runner."""

import os
import unittest

from email import message_from_file
from mailman.app.lifecycle import create_list
from mailman.config import config
from mailman.interfaces.archiver import IArchiver
from mailman.interfaces.mailinglist import IListArchiverSet
from mailman.runners.archive import ArchiveRunner
from mailman.testing.helpers import (
    configuration,
    get_queue_messages,
    LogFileMark,
    make_testable_runner,
    specialized_message_from_string as mfs,
)
from mailman.testing.layers import ConfigLayer
from mailman.utilities.datetime import factory, now, RFC822_DATE_FMT
from zope.interface import implementer


@implementer(IArchiver)
class DummyArchiver:
    name = 'dummy'

    @staticmethod
    def list_url(mlist):
        return 'http://archive.example.com/'

    @staticmethod
    def permalink(mlist, msg):
        filename = msg['message-id-hash']
        return 'http://archive.example.com/' + filename

    @staticmethod
    def archive_message(mlist, msg):
        filename = msg['message-id-hash']
        path = os.path.join(config.MESSAGES_DIR, filename)
        with open(path, 'w') as fp:
            print(msg.as_string(), file=fp)
        # Not technically allowed by the API, but good enough for the test.
        return path


@implementer(IArchiver)
class BrokenArchiver:
    """An archiver that has some broken methods."""

    name = 'broken'

    def list_url(self, mlist):
        raise RuntimeError('Cannot get list URL')

    def permalink(self, mlist, msg):
        raise RuntimeError('Cannot get permalink')

    @staticmethod
    def archive_message(mlist, message):
        raise RuntimeError('Cannot archive message')


class TestArchiveRunner(unittest.TestCase):
    """Test the archive runner."""

    layer = ConfigLayer

    def setUp(self):
        self._mlist = create_list('test@example.com')
        self._now = now()
        # Enable just the dummy archiver.
        config.push('dummy', """
        [archiver.dummy]
        class: mailman.runners.tests.test_archiver.DummyArchiver
        enable: no
        [archiver.broken]
        class: mailman.runners.tests.test_archiver.BrokenArchiver
        enable: no
        [archiver.prototype]
        enable: no
        [archiver.mhonarc]
        enable: no
        [archiver.mail_archive]
        enable: no
        """)
        self.addCleanup(config.pop, 'dummy')
        self._archiveq = config.switchboards['archive']
        self._msg = mfs("""\
From: aperson@example.com
To: test@example.com
Subject: My first post
Message-ID: <first>
Message-ID-Hash: 4CMWUN6BHVCMHMDAOSJZ2Q72G5M32MWB

First post!
""")
        self._runner = make_testable_runner(ArchiveRunner)
        IListArchiverSet(self._mlist).get('dummy').is_enabled = True

    @configuration('archiver.dummy', enable='yes')
    def test_archive_runner(self):
        # Ensure that the archive runner ends up archiving the message.
        self._archiveq.enqueue(
            self._msg, {},
            listid=self._mlist.list_id,
            received_time=now())
        self._runner.run()
        # There should now be a copy of the message in the file system.
        filename = os.path.join(
            config.MESSAGES_DIR, '4CMWUN6BHVCMHMDAOSJZ2Q72G5M32MWB')
        with open(filename) as fp:
            archived = message_from_file(fp)
        self.assertEqual(archived['message-id'], '<first>')

    @configuration('archiver.dummy', enable='yes')
    def test_archive_runner_with_dated_message(self):
        # Date headers don't throw off the archiver runner.
        self._msg['Date'] = now(strip_tzinfo=False).strftime(RFC822_DATE_FMT)
        self._archiveq.enqueue(
            self._msg, {},
            listid=self._mlist.list_id,
            received_time=now())
        self._runner.run()
        # There should now be a copy of the message in the file system.
        filename = os.path.join(
            config.MESSAGES_DIR, '4CMWUN6BHVCMHMDAOSJZ2Q72G5M32MWB')
        with open(filename) as fp:
            archived = message_from_file(fp)
        self.assertEqual(archived['message-id'], '<first>')
        self.assertEqual(archived['date'], 'Mon, 01 Aug 2005 07:49:23 +0000')

    @configuration('archiver.dummy', enable='yes', clobber_date='never')
    def test_clobber_date_never(self):
        # Even if the Date header is insanely off from the received time of
        # the message, if clobber_date is 'never', the header is not clobbered.
        self._msg['Date'] = now(strip_tzinfo=False).strftime(RFC822_DATE_FMT)
        self._archiveq.enqueue(
            self._msg, {},
            listid=self._mlist.list_id,
            received_time=now())
        self._runner.run()
        # There should now be a copy of the message in the file system.
        filename = os.path.join(
            config.MESSAGES_DIR, '4CMWUN6BHVCMHMDAOSJZ2Q72G5M32MWB')
        with open(filename) as fp:
            archived = message_from_file(fp)
        self.assertEqual(archived['message-id'], '<first>')
        self.assertEqual(archived['date'], 'Mon, 01 Aug 2005 07:49:23 +0000')

    @configuration('archiver.dummy', enable='yes')
    def test_clobber_dateless(self):
        # A message with no Date header will always get clobbered.
        self.assertEqual(self._msg['date'], None)
        # Now, before enqueuing the message (well, really, calling 'now()'
        # again), fast forward a few days.
        self._archiveq.enqueue(
            self._msg, {},
            listid=self._mlist.list_id,
            received_time=now(strip_tzinfo=False))
        self._runner.run()
        # There should now be a copy of the message in the file system.
        filename = os.path.join(
            config.MESSAGES_DIR, '4CMWUN6BHVCMHMDAOSJZ2Q72G5M32MWB')
        with open(filename) as fp:
            archived = message_from_file(fp)
        self.assertEqual(archived['message-id'], '<first>')
        self.assertEqual(archived['date'], 'Mon, 01 Aug 2005 07:49:23 +0000')

    @configuration('archiver.dummy', enable='yes', clobber_date='always')
    def test_clobber_date_always(self):
        # The date always gets clobbered with the current received time.
        self._msg['Date'] = now(strip_tzinfo=False).strftime(RFC822_DATE_FMT)
        # Now, before enqueuing the message (well, really, calling 'now()'
        # again as will happen in the runner), fast forward a few days.
        self._archiveq.enqueue(
            self._msg, {},
            listid=self._mlist.list_id)
        factory.fast_forward(days=4)
        self._runner.run()
        # There should now be a copy of the message in the file system.
        filename = os.path.join(
            config.MESSAGES_DIR, '4CMWUN6BHVCMHMDAOSJZ2Q72G5M32MWB')
        with open(filename) as fp:
            archived = message_from_file(fp)
        self.assertEqual(archived['message-id'], '<first>')
        self.assertEqual(archived['date'], 'Fri, 05 Aug 2005 07:49:23 +0000')
        self.assertEqual(archived['x-original-date'],
                         'Mon, 01 Aug 2005 07:49:23 +0000')

    @configuration('archiver.dummy',
                   enable='yes', clobber_date='maybe', clobber_skew='1d')
    def test_clobber_date_maybe_when_insane(self):
        # The date is clobbered if it's farther off from now than its skew
        # period.
        self._msg['Date'] = now(strip_tzinfo=False).strftime(RFC822_DATE_FMT)
        # Now, before enqueuing the message (well, really, calling 'now()'
        # again as will happen in the runner), fast forward a few days.
        self._archiveq.enqueue(
            self._msg, {},
            listid=self._mlist.list_id)
        factory.fast_forward(days=4)
        self._runner.run()
        # There should now be a copy of the message in the file system.
        filename = os.path.join(
            config.MESSAGES_DIR, '4CMWUN6BHVCMHMDAOSJZ2Q72G5M32MWB')
        with open(filename) as fp:
            archived = message_from_file(fp)
        self.assertEqual(archived['message-id'], '<first>')
        self.assertEqual(archived['date'], 'Fri, 05 Aug 2005 07:49:23 +0000')
        self.assertEqual(archived['x-original-date'],
                         'Mon, 01 Aug 2005 07:49:23 +0000')

    @configuration('archiver.dummy',
                   enable='yes', clobber_date='maybe', clobber_skew='10d')
    def test_clobber_date_maybe_when_sane(self):
        # The date is not clobbered if it's nearer to now than its skew
        # period.
        self._msg['Date'] = now(strip_tzinfo=False).strftime(RFC822_DATE_FMT)
        # Now, before enqueuing the message (well, really, calling 'now()'
        # again as will happen in the runner), fast forward a few days.
        self._archiveq.enqueue(
            self._msg, {},
            listid=self._mlist.list_id)
        factory.fast_forward(days=4)
        self._runner.run()
        # There should now be a copy of the message in the file system.
        filename = os.path.join(
            config.MESSAGES_DIR, '4CMWUN6BHVCMHMDAOSJZ2Q72G5M32MWB')
        with open(filename) as fp:
            archived = message_from_file(fp)
        self.assertEqual(archived['message-id'], '<first>')
        self.assertEqual(archived['date'], 'Mon, 01 Aug 2005 07:49:23 +0000')
        self.assertEqual(archived['x-original-date'], None)

    @configuration('archiver.dummy', enable='yes')
    def test_disable_all_list_archivers(self):
        # Let's disable all the archivers for the mailing list, but not the
        # global archivers.  No messages will get archived.
        for archiver in IListArchiverSet(self._mlist).archivers:
            archiver.is_enabled = False
        config.db.store.commit()
        self._archiveq.enqueue(
            self._msg, {},
            listid=self._mlist.list_id)
        self._runner.run()
        self.assertEqual(os.listdir(config.MESSAGES_DIR), [])

    @configuration('archiver.broken', enable='yes')
    def test_broken_archiver(self):
        # GL issue #208 - IArchive messages raise exceptions, breaking the
        # rfc-2369 handler and shunting messages.
        mark = LogFileMark('mailman.archiver')
        self._archiveq.enqueue(
            self._msg, {},
            listid=self._mlist.list_id,
            received_time=now())
        IListArchiverSet(self._mlist).get('broken').is_enabled = True
        self._runner.run()
        # The archiver is broken, so there are no messages on the file system,
        # but there is a log message and the message was not shunted.
        log_messages = mark.read()
        self.assertIn('Exception in "broken" archiver', log_messages)
        self.assertIn('RuntimeError: Cannot archive message', log_messages)
        get_queue_messages('shunt', expected_count=0)



# Copyright (C) 2008-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Various test helpers."""

import os
import sys
import time
import uuid
import shutil
import signal
import socket
import logging
import smtplib
import datetime
import threading

from contextlib import contextmanager, suppress
from email import message_from_bytes, message_from_string
from lazr.config import as_timedelta
from mailman.bin.master import Loop as Master
from mailman.config import config
from mailman.database.transaction import transaction
from mailman.email.message import Message
from mailman.interfaces.action import Action
from mailman.interfaces.member import DeliveryStatus, MemberRole
from mailman.interfaces.messages import IMessageStore
from mailman.interfaces.styles import IStyleManager
from mailman.interfaces.usermanager import IUserManager
from mailman.runners.digest import DigestRunner
from mailman.utilities.mailbox import Mailbox
from public import public
from requests import request
from unittest import mock
from urllib.error import HTTPError
from zope import event
from zope.component import getUtility


NL = '\n'


@public
def make_testable_runner(runner_class, name=None, predicate=None):
    """Create a runner that runs until its queue is empty.

    :param runner_class: The runner class.
    :type runner_class: class
    :param name: Optional queue name; if not given, it is calculated from the
        class name.
    :type name: string or None
    :param predicate: Optional alternative predicate for deciding when to stop
        the runner.  When None (the default) it stops when the queue is empty.
    :type predicate: callable that gets one argument, the queue runner.
    :return: A runner instance.
    """
    if name is None:
        assert runner_class.__name__.endswith('Runner'), (
            'Unparseable runner class name: %s' % runner_class.__name__)
        name = runner_class.__name__[:-6].lower()

    class EmptyingRunner(runner_class):
        """Stop processing when the queue is empty."""

        def __init__(self, *args, **kws):
            super().__init__(*args, **kws)
            # We know it's an EmptyingRunner, so really we want to see the
            # super class in the log files.
            self.__class__.__name__ = runner_class.__name__

        def _do_periodic(self):
            """Stop when the queue is empty."""
            super()._do_periodic()
            if predicate is None:
                if self.switchboard is None:
                    self._stop = True
                else:
                    self._stop = (len(self.switchboard.files) == 0)
            else:
                self._stop = predicate(self)

    return EmptyingRunner(name)


class _Bag:
    def __init__(self, **kws):
        for key, value in kws.items():
            setattr(self, key, value)


@public
def get_queue_messages(queue_name, sort_on=None, expected_count=None):
    """Return and clear all the messages in the given queue.

    :param queue_name: A string naming a queue.
    :param sort_on: The message header to sort on.  If None (the default),
        no sorting is performed.
    :param expected_count: If given and there aren't exactly this number of
        messages in the queue, raise an AssertionError.
    :return: A list of 2-tuples where each item contains the message and
        message metadata.
    """
    queue = config.switchboards[queue_name]
    messages = []
    for filebase in queue.files:
        msg, msgdata = queue.dequeue(filebase)
        messages.append(_Bag(msg=msg, msgdata=msgdata))
        queue.finish(filebase)
    if expected_count is not None:
        if len(messages) != expected_count:
            for item in messages:
                print(item.msg, file=sys.stderr)
            raise AssertionError('Wanted {}, got {}'.format(
                expected_count, len(messages)))
    if sort_on is not None:
        messages.sort(key=lambda item: str(item.msg[sort_on]))
    return messages


@public
def digest_mbox(mlist):
    """The mailing list's pending digest as a mailbox.

    :param mlist: The mailing list.
    :return: The mailing list's pending digest as a mailbox.
    """
    path = os.path.join(mlist.data_path, 'digest.mmdf')
    return Mailbox(path)


# Remember, Master is mailman.bin.master.Loop.
@public
class TestableMaster(Master):
    """A testable master loop watcher."""

    def __init__(self, start_check=None):
        """Create a testable master loop watcher.

        :param start_check: Optional callable used to check whether everything
            is running as the test expects.  Called in `loop()` in the
            subthread before the event is set.  The callback should block
            until the pass condition is set.
        :type start_check: Callable taking no arguments, returning nothing.
        """
        super().__init__(
            restartable=False,
            config_file=os.environ.get('MAILMAN_CONFIG_FILE', config.filename))
        self.start_check = start_check
        self.event = threading.Event()
        self.thread = threading.Thread(target=self.loop)
        self.thread.daemon = True
        self._started_kids = None

    def _pause(self):
        """See `Master`."""
        # No-op this because the tests generally do not signal the master,
        # which would mean the signal.pause() never exits.
        pass

    def start(self, *runners):
        """Start the master."""
        self.start_runners(runners)
        self.thread.start()
        # Wait until all the children are definitely started.
        self.event.wait()

    def stop(self):
        """Stop the master by killing all the children."""
        for pid in self.runner_pids:
            os.kill(pid, signal.SIGTERM)
        self.cleanup()
        self.thread.join()

    def loop(self):
        """Wait until all the runners are actually running before looping."""
        starting_kids = set(self._kids)
        while starting_kids:
            for pid in self._kids:
                # Ignore the exception which gets raised when the child has
                # not yet started.
                with suppress(ProcessLookupError):
                    os.kill(pid, 0)
                    starting_kids.remove(pid)
        # Keeping a copy of all the started child processes for use by the
        # testing environment, even after all have exited.
        self._started_kids = set(self._kids)
        # If there are extra conditions to check, do it now.
        if self.start_check is not None:
            self.start_check()
        # Let the blocking thread know everything's running.
        self.event.set()
        super().loop()

    @property
    def runner_pids(self):
        """The pids of all the child runner processes."""
        yield from self._started_kids


class LMTP(smtplib.SMTP):
    """Like a normal SMTP client, but for LMTP."""
    def lhlo(self, name=''):
        self.putcmd('lhlo', name or self.local_hostname)
        code, msg = self.getreply()
        self.helo_resp = msg
        return code, msg


@public
def get_lmtp_client(quiet=False):
    """Return a connected LMTP client."""
    # It's possible the process has started but is not yet accepting
    # connections.  Wait a little while.
    lmtp = LMTP()
    # lmtp.debuglevel = 1
    until = datetime.datetime.now() + as_timedelta(config.devmode.wait)
    while datetime.datetime.now() < until:
        try:
            response = lmtp.connect(
                config.mta.lmtp_host, int(config.mta.lmtp_port))
            if not quiet:
                print(response)
            return lmtp
        except ConnectionRefusedError:
            time.sleep(0.1)
    else:
        raise RuntimeError('Connection refused')


@public
def get_nntp_server(cleanups):
    """Create and start an NNTP server mock.

    This can be used to retrieve the posted message for verification.
    """
    patcher = mock.patch('nntplib.NNTP')
    server_class = patcher.start()
    cleanups.callback(patcher.stop)
    nntpd = server_class()
    # A class for more convenient access to the posted message.
    class NNTPProxy:                                              # noqa: E306
        def get_message(self):
            args = nntpd.post.call_args
            return message_from_bytes(args[0][0].read())
    return NNTPProxy()


@public
def wait_for_webservice(hostname=None, port=None):
    """Wait for the REST server to start serving requests."""
    hostname = config.webservice.hostname if hostname is None else hostname
    port = int(config.webservice.port) if port is None else port
    until = datetime.datetime.now() + as_timedelta(config.devmode.wait)
    while datetime.datetime.now() < until:
        try:
            socket.socket().connect((hostname, port))
        except ConnectionRefusedError:
            time.sleep(0.1)
        else:
            break
    else:
        raise RuntimeError('Connection refused')


@public
def call_api(url, data=None, method=None, username=None, password=None,
             headers=None, json=None):
    """'Call a URL with a given HTTP method and return the resulting object.

    The object will have been JSON decoded.

    :param url: The url to open, read, and print.
    :type url: string
    :param data: Data to use to POST to a URL.
    :type data: dict
    :param method: Alternative HTTP method to use.
    :type method: str
    :param username: The HTTP Basic Auth user name.  None means use the value
        from the configuration.
    :type username: str
    :param password: The HTTP Basic Auth password.  None means use the value
        from the configuration.
    :type password: str
    :param headers: List of additional headers to be included with the request.
    :type headers: A dictionary of {'Header': 'Value'}
    :param json: JSON body for the request.
    :type json: dict
    :return: A 2-tuple containing the JSON decoded content (if there is any,
        else None) and the response object.
    :rtype: 2-tuple of (dict, response)
    :raises HTTPError: when a non-2xx return code is received.
    """
    if method is None:
        if data is None and json is None:
            method = 'GET'
        else:
            method = 'POST'
    method = method.upper()
    basic_auth = (
        (config.webservice.admin_user if username is None else username),
        (config.webservice.admin_pass if password is None else password))
    response = request(
        method, url, data=data, auth=basic_auth, headers=headers, json=json)
    # For backward compatibility with existing doctests, turn non-2xx response
    # codes into a urllib.error exceptions.
    if response.status_code // 100 != 2:
        content = None
        content_type = response.headers.get('content-type')
        if ('application/json' in content_type and
                response.content):
            content = response.json().get('description', None)
        else:
            content = response.text
        raise HTTPError(
            url, response.status_code, content, response, None)
    if len(response.content) == 0:
        return None, response
    return response.json(), response


@public
@contextmanager
def event_subscribers(*subscribers):
    """Temporarily extend the Zope event subscribers list.

    :param subscribers: A sequence of event subscribers.
    :type subscribers: sequence of callables, each receiving one argument, the
        event.
    """
    old_subscribers = event.subscribers[:]
    event.subscribers.extend(subscribers)
    try:
        yield
    finally:
        event.subscribers[:] = old_subscribers


@public
class configuration:
    """A decorator/context manager for temporarily setting configurations."""

    def __init__(self, section, **kws):
        self._section = section
        # Most tests don't care about the name given to the temporary
        # configuration.  Usually we'll just craft a random one, but some
        # tests do care, so give them a hook to set it.
        if '_configname' in kws:
            self._uuid = kws.pop('_configname')
        else:
            self._uuid = uuid.uuid4().hex
        self._values = kws.copy()

    def _apply(self):
        lines = ['[{}]'.format(self._section)]
        for key, value in self._values.items():
            lines.append('{}: {}'.format(key, value))
        config.push(self._uuid, NL.join(lines))

    def _remove(self):
        config.pop(self._uuid)

    def __enter__(self):
        self._apply()
        return self

    def __exit__(self, *exc_info):
        self._remove()
        # Do not suppress exceptions.
        return False

    def __call__(self, func):
        def wrapper(*args, **kws):
            self._apply()
            try:
                return func(*args, **kws)
            finally:
                self._remove()
        return wrapper


@public
@contextmanager
def temporary_db(db):
    real_db = config.db
    config.db = db
    try:
        yield
    finally:
        config.db = real_db


@public
class chdir:
    """A context manager for temporary directory changing."""
    def __init__(self, directory):
        self._curdir = None
        self._directory = directory

    def __enter__(self):
        self._curdir = os.getcwd()
        os.chdir(self._directory)

    def __exit__(self, exc_type, exc_val, exc_tb):
        os.chdir(self._curdir)
        # Don't suppress exceptions.
        return False


@public
def subscribe(mlist, first_name, role=MemberRole.member, email=None,
              as_user=False):
    """Helper for subscribing a sample person to a mailing list.

    Returns the newly created member object.
    """
    user_manager = getUtility(IUserManager)
    email = ('{}person@example.com'.format(first_name[0].lower())
             if email is None else email)
    full_name = '{} Person'.format(first_name)
    with transaction():
        person = user_manager.get_user(email)
        if person is None:
            address = user_manager.get_address(email)
            if address is None:
                person = user_manager.create_user(email, full_name)
                set_preferred(person)
                if as_user:
                    subscription_object = person
                else:
                    subscription_object = person.preferred_address
            else:
                if not as_user:
                    subscription_object = address
                else:
                    person = user_manager.create_user(display_name=full_name)
                    person.link(address)
                    set_preferred(person)
                    subscription_object = person
        else:
            if as_user:
                if person.preferred_address is None:
                    set_preferred(person)
                subscription_object = person
            else:
                subscription_object = list(person.addresses)[0]
        mlist.subscribe(subscription_object, role)
        roster = mlist.get_roster(role)
        return roster.get_member(email)


@public
def reset_the_world():
    """Reset everything:

    * Clear out the database
    * Remove all residual queue and digest files
    * Clear the message store
    * Reset the global style manager

    This should be as thorough a reset of the system as necessary to keep
    tests isolated.
    """
    # Reset the database between tests.
    config.db._reset()
    # Remove any digest files and members.txt file (for the file-recips
    # handler) in the lists' data directories.
    for dirpath, dirnames, filenames in os.walk(config.LIST_DATA_DIR):
        for filename in filenames:
            if filename.endswith('.mmdf') or filename == 'members.txt':
                os.remove(os.path.join(dirpath, filename))
    # Remove all residual queue files.
    for dirpath, dirnames, filenames in os.walk(config.QUEUE_DIR):
        for filename in filenames:
            os.remove(os.path.join(dirpath, filename))
    # Clear out messages in the message store.
    message_store = getUtility(IMessageStore)
    with transaction():
        for message in message_store.messages:
            message_store.delete_message(message['message-id'])
    # Delete any other residual messages.
    for dirpath, dirnames, filenames in os.walk(config.MESSAGES_DIR):
        for filename in filenames:
            os.remove(os.path.join(dirpath, filename))
        shutil.rmtree(dirpath)
    # Remove all the cache subdirectories, recursively.
    for dirname in os.listdir(config.CACHE_DIR):
        shutil.rmtree(os.path.join(config.CACHE_DIR, dirname))
    # Reset the global style manager.
    getUtility(IStyleManager).populate()
    # Remove all dynamic header-match rules.
    config.chains['header-match'].flush()
    # Remove cached organizational domain suffix file.
    from mailman.rules.dmarc import LOCAL_FILE_NAME
    suffix_file = os.path.join(config.VAR_DIR, LOCAL_FILE_NAME)
    with suppress(FileNotFoundError):
        os.remove(suffix_file)


@public
def specialized_message_from_string(unicode_text):
    """Parse text into a message object.

    This is specialized in the sense that an instance of Mailman's own Message
    object is returned, and this message object has an attribute
    `original_size` which is the pre-calculated size in bytes of the message's
    text representation.

    Also, the text must be ASCII-only unicode.
    """
    # This mimic what Switchboard.dequeue() does when parsing a message from
    # text into a Message instance.
    original_size = len(unicode_text)
    message = message_from_string(unicode_text, Message)
    message.original_size = original_size
    return message


@public
class LogFileMark:
    def __init__(self, log_name):
        self._log = logging.getLogger(log_name)
        self._filename = self._log.handlers[0].filename
        self._filepos = os.stat(self._filename).st_size

    def readline(self):
        with open(self._filename) as fp:
            fp.seek(self._filepos)
            return fp.readline()

    def read(self):
        with open(self._filename) as fp:
            fp.seek(self._filepos)
            return fp.read()


@public
def make_digest_messages(mlist, msg=None):
    if msg is None:
        msg = specialized_message_from_string("""\
From: anne@example.org
To: {listname}
Message-ID: <testing>

message triggering a digest
""".format(listname=mlist.fqdn_listname))
    mbox_path = os.path.join(mlist.data_path, 'digest.mmdf')
    config.handlers['to-digest'].process(mlist, msg, {})
    config.switchboards['digest'].enqueue(
        msg,
        listname=mlist.fqdn_listname,
        digest_path=mbox_path,
        volume=1, digest_number=1)
    runner = make_testable_runner(DigestRunner, 'digest')
    runner.run()


@public
def set_preferred(user):
    # Avoid circular imports.
    from mailman.utilities.datetime import now
    preferred = list(user.addresses)[0]
    preferred.verified_on = now()
    user.preferred_address = preferred
    return preferred


@public
@contextmanager
def hackenv(envar, new_value):
    """Hack the environment temporarily, then reset it."""
    old_value = os.getenv(envar)
    if new_value is None:
        if envar in os.environ:
            del os.environ[envar]
    else:
        os.environ[envar] = new_value
    try:
        yield
    finally:
        if old_value is None:
            if envar in os.environ:
                del os.environ[envar]
        else:
            os.environ[envar] = old_value


def nose2_start_test_run_callback(plugin):
    from mailman.testing.layers import ConfigLayer, MockAndMonkeyLayer
    MockAndMonkeyLayer.testing_mode = True
    if (plugin.stderr or
            len(os.environ.get('MM_VERBOSE_TESTLOG', '').strip()) > 0):
        ConfigLayer.stderr = True


def set_moderation(mlist, subscriber, action):
    """Set moderation action for the subscriber of mailing list."""
    member = mlist.members.get_member(subscriber)
    member.moderation_action = Action[action]


def set_delivery(mlist, subscriber, delivery_status):
    """Set delivery_status to DeliveryStatus"""
    member = mlist.members.get_member(subscriber)
    member.preferences.delivery_status = DeliveryStatus[delivery_status]



# Copyright (C) 2008-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Base class for terminal chains."""

from mailman.config import config
from mailman.interfaces.chain import (
    IChain,
    IChainIterator,
    IChainLink,
    IMutableChain,
    LinkAction,
)
from mailman.interfaces.rules import IRule
from mailman.utilities.modules import abstract_component
from public import public
from zope.interface import implementer


@public
@implementer(IChainLink)
class Link:
    """A chain link."""

    def __init__(self, rule, action=None, chain=None, function=None):
        self.rule = (rule
                     if IRule.providedBy(rule)
                     else config.rules[rule])
        self.action = (LinkAction.defer if action is None else action)
        self.chain = (chain
                      if chain is None or IChain.providedBy(chain)
                      else config.chains[chain])
        self.function = function

    def __repr__(self):
        message = '<Link "if {0.rule.name} then {0.action}"'
        if self.chain is None and self.function is not None:
            message += ' {0.function.__name__}()'
        elif self.chain is not None and self.function is None:
            message += ' {0.chain.name}'
        elif self.chain is None and self.function is None:
            pass
        else:
            message += ' {0.chain.name} {0.function.__name__}()'
        message += '>'
        return message.format(self)


@public
@abstract_component
@implementer(IChain, IChainIterator)
class TerminalChainBase:
    """A base chain that always matches and executes a method.

    The method is called '_process()' and must be provided by the subclass.
    """
    def _process(self, mlist, msg, msgdata):
        """Process the message for the given mailing list.

        This must be overridden by subclasses.

        :param mlist: The mailing list.
        :param msg: The message.
        :param msgdata: The message metadata.
        """
        raise NotImplementedError

    def get_links(self, mlist, msg, msgdata):
        """See `IChain`."""
        return iter(self)

    def __iter__(self):
        """See `IChainIterator`."""
        # First, yield a link that always runs the process method.
        yield Link('truth', LinkAction.run, function=self._process)
        # Now yield a rule that stops all processing.
        yield Link('truth', LinkAction.stop)


@public
@abstract_component
@implementer(IChain)
class JumpChainBase:
    """A base chain that simplifies jumping to another chain."""
    def jump_to(self, mlist, msg, msgsdata):
        """Return the chain to jump to.

        This must be overridden by subclasses.

        :param mlist: The mailing list.
        :param msg: The message.
        :param msgdata: The message metadata.
        :return: The name of the chain to jump to.
        :rtype: str
        """
        raise NotImplementedError

    def get_links(self, mlist, msg, msgdata):
        jump_chain = self.jump_to(mlist, msg, msgdata)
        return iter([
            Link('truth', LinkAction.jump, jump_chain),
            ])


@public
@abstract_component
@implementer(IMutableChain)
class Chain:
    """Generic chain base class."""

    def __init__(self, name, description):
        assert name not in config.chains, (
            'Duplicate chain name: {}'.format(name))
        self.name = name
        self.description = description
        self._links = []

    def append_link(self, link):
        """See `IMutableChain`."""
        self._links.append(link)

    def flush(self):
        """See `IMutableChain`."""
        self._links = []

    def get_links(self, mlist, msg, msgdata):
        """See `IChain`."""
        return iter(ChainIterator(self))

    def get_iterator(self):
        """Return an iterator over the links."""
        # We do it this way in order to preserve a separation of interfaces,
        # and allows .get_links() to be overridden.
        yield from self._links


@public
@implementer(IChainIterator)
class ChainIterator:
    """Generic chain iterator."""

    def __init__(self, chain):
        self._chain = chain

    def __iter__(self):
        """See `IChainIterator`."""
        return self._chain.get_iterator()



# Copyright (C) 2007-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Model for users."""

from mailman.database.model import Model
from mailman.database.transaction import dbconnection
from mailman.database.types import SAUnicode4Byte, UUID
from mailman.interfaces.address import (
    AddressAlreadyLinkedError,
    AddressNotLinkedError,
)
from mailman.interfaces.user import (
    IUser,
    PasswordChangeEvent,
    UnverifiedAddressError,
)
from mailman.model.address import Address
from mailman.model.member import Member
from mailman.model.preferences import Preferences
from mailman.model.roster import Memberships
from mailman.utilities.datetime import factory as date_factory
from mailman.utilities.uid import UIDFactory
from public import public
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, Integer
from sqlalchemy.orm import backref, relationship
from zope.event import notify
from zope.interface import implementer


uid_factory = UIDFactory(context='users')


@public
@implementer(IUser)
class User(Model):
    """Mailman users."""

    __tablename__ = 'user'

    id = Column(Integer, primary_key=True)
    display_name = Column(SAUnicode4Byte)
    _password = Column('password', SAUnicode4Byte)
    _user_id = Column(UUID, index=True)
    _created_on = Column(DateTime)
    is_server_owner = Column(Boolean, default=False)

    addresses = relationship(
        'Address', backref='user',
        primaryjoin=(id == Address.user_id))

    _preferred_address_id = Column(
        Integer,
        ForeignKey('address.id', use_alter=True,
                   name='_preferred_address',
                   ondelete='SET NULL'))

    _preferred_address = relationship(
        'Address', primaryjoin=(_preferred_address_id == Address.id),
        post_update=True)

    preferences_id = Column(Integer, ForeignKey('preferences.id'), index=True)
    preferences = relationship(
        'Preferences', backref=backref('user', uselist=False))

    @dbconnection
    def __init__(self, store, display_name=None, preferences=None):
        super().__init__()
        self._created_on = date_factory.now()
        user_id = uid_factory.new()
        assert store.query(User).filter_by(_user_id=user_id).count() == 0, (
            'Duplicate user id {}'.format(user_id))
        self._user_id = user_id
        self.display_name = ('' if display_name is None else display_name)
        if preferences is not None:
            store.add(preferences)
            self.preferences = preferences
        store.add(self)

    def __repr__(self):
        short_user_id = self.user_id.int
        return '<User "{0.display_name}" ({2}) at {1:#x}>'.format(
            self, id(self), short_user_id)

    @property
    def user_id(self):
        """See `IUser`."""
        return self._user_id

    @property
    def created_on(self):
        """See `IUser`."""
        return self._created_on

    @property
    def password(self):
        return self._password

    @password.setter
    def password(self, new_password):
        self._password = new_password
        notify(PasswordChangeEvent(self))

    def link(self, address):
        """See `IUser`."""
        if address.user is not None:
            raise AddressAlreadyLinkedError(address)
        address.user = self

    def unlink(self, address):
        """See `IUser`."""
        if address.user is None or address.user is not self:
            raise AddressNotLinkedError(address)
        address.user = None

    @property
    def preferred_address(self):
        """See `IUser`."""
        return self._preferred_address

    @preferred_address.setter
    def preferred_address(self, address):
        """See `IUser`."""
        if address.verified_on is None:
            raise UnverifiedAddressError(
                "{} must be verified before setting as primary".format(address)
                    )
        if self.controls(address.email):
            # This user already controls the email address.
            pass
        elif address.user is None:
            self.link(address)
        elif address.user != self:
            raise AddressAlreadyLinkedError(address)
        self._preferred_address = address

    @preferred_address.deleter
    def preferred_address(self):
        """See `IUser`."""
        self._preferred_address = None

    @dbconnection
    def controls(self, store, email):
        """See `IUser`."""
        found = store.query(Address).filter_by(email=email)
        if found.count() == 0:
            return False
        assert found.count() == 1, 'Unexpected count'
        return found[0].user is self

    @dbconnection
    def register(self, store, email, display_name=None):
        """See `IUser`."""
        # First, see if the address already exists
        address = store.query(Address).filter_by(email=email.lower()).first()
        if address is None:
            if display_name is None:
                display_name = ''
            address = Address(email=email, display_name=display_name)
            address.preferences = Preferences()
        # Link the address to the user if it is not already linked.
        if address.user is not None:
            raise AddressAlreadyLinkedError(address)
        address.user = self
        store.add(address)
        return address

    @property
    def memberships(self):
        return Memberships(self)

    @dbconnection
    def absorb(self, store, user):
        """See `IUser`."""
        if not isinstance(user, User):
            raise TypeError('Not a user {!r}'.format(user))
        if user.id == self.id:
            # It's a no-op to absorb ourself.
            return
        # Relink addresses.
        for address in list(user.addresses):
            # Convert these to a list because we'll mutate the result.
            address.user = self
        # Merge memberships.
        other_members = store.query(Member).filter(
            Member.user_id == user.id)
        my_subscriptions = set(
            (member.list_id, member.role)
            for member in self.memberships.members)
        for member in other_members:
            # Only import memberships for list/roles I'm not already a member
            # with.  This prevents duplicate memberships.
            if (member.list_id, member.role) not in my_subscriptions:
                member.user_id = self.id
            else:
                store.delete(member)
        # Merge the user preferences.
        self.preferences.absorb(user.preferences)
        store.delete(user.preferences)
        # Merge display_name, password and is_server_owner attributes.
        for name in ('display_name', 'password', 'is_server_owner'):
            if getattr(user, name) and not getattr(self, name):
                setattr(self, name, getattr(user, name))
        # Delete the other user.
        store.delete(user)


@public
class DomainOwner(Model):
    """Internal table for associating domains to their owners."""

    __tablename__ = 'domain_owner'

    user_id = Column(Integer, ForeignKey('user.id'), primary_key=True)
    domain_id = Column(Integer, ForeignKey('domain.id'), primary_key=True)



# Copyright (C) 2006-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Common database support."""

import logging

from mailman.config import config
from mailman.interfaces.database import IDatabase
from mailman.utilities.string import expand
from public import public
from sqlalchemy import create_engine
from sqlalchemy.exc import PendingRollbackError
from sqlalchemy.orm import scoped_session, sessionmaker
from zope.interface import implementer


log = logging.getLogger('mailman.database')


@public
@implementer(IDatabase)
class SABaseDatabase:
    """The database base class for use with SQLAlchemy.

    Use this as a base class for your DB-Specific derived classes.
    """
    def __init__(self):
        self.url = None
        self.pool_class = None

    def begin(self):
        """See `IDatabase`."""
        # SQLAlchemy does this for us.
        pass

    def commit(self):
        """See `IDatabase`."""
        try:
            self.store.commit()
        except PendingRollbackError:
            self.abort()                    # pragma: nocover

    def abort(self):
        """See `IDatabase`."""
        self.store.rollback()

    def _pre_reset(self, store):
        """Clean up method for testing.

        This method is called during the test suite just before all the model
        tables are removed.  Override this to perform any database-specific
        pre-removal cleanup.
        """
        pass

    def _post_reset(self, store):
        """Clean up method for testing.

        This method is called during the test suite just after all the model
        tables have been removed.  Override this to perform any
        database-specific post-removal cleanup.
        """
        pass

    def _prepare(self, url):
        """Prepare the database for creation.

        Some database backends need to do so me prep work before letting Storm
        create the database.  For example, we have to touch the SQLite .db
        file first so that it has the proper file modes.
        """
        pass

    @property
    def store(self):
        return self.sessionmaker()

    def close_session(self):
        self.store.close()

    def initialize(self, debug=None):
        """See `IDatabase`."""
        # Calculate the engine url.
        url = expand(config.database.url, None, config.paths)
        self._prepare(url)
        log.debug('Database url: %s', url)
        # XXX By design of SQLite, database file creation does not honor
        # umask.  See their ticket #1193:
        # http://www.sqlite.org/cvstrac/tktview?tn=1193,31
        #
        # This sucks for us because the mailman.db file /must/ be group
        # writable, however even though we guarantee our umask is 002 here, it
        # still gets created without the necessary g+w permission, due to
        # SQLite's policy.  This should only affect SQLite engines because its
        # the only one that creates a little file on the local file system.
        # This kludges around their bug by "touch"ing the database file before
        # SQLite has any chance to create it, thus honoring the umask and
        # ensuring the right permissions.  We only try to do this for SQLite
        # engines, and yes, we could have chmod'd the file after the fact, but
        # half dozen and all...
        self.url = url
        self.engine = create_engine(
            url,
            isolation_level='READ UNCOMMITTED',
            pool_pre_ping=True,
            poolclass=self.pool_class,
            future=True,
            )
        self.sessionmaker = scoped_session(sessionmaker(
            bind=self.engine, expire_on_commit=True, future=True))



# Copyright (C) 2007-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Interface for a roster of members."""

from public import public
from zope.interface import Attribute, Interface


@public
class IRoster(Interface):
    """A roster is a collection of `IMembers`."""

    name = Attribute(
        """The name for this roster.

        Rosters are considered equal if they have the same name.""")

    members = Attribute(
        """An iterator over all the IMembers managed by this roster.""")

    member_count = Attribute(
        """The number of members managed by this roster.""")

    users = Attribute(
        """An iterator over all the IUsers reachable by this roster.

        This returns all the users for all the members managed by this roster.
        """)

    addresses = Attribute(
        """An iterator over all the IAddresses reachable by this roster.

        This returns all the addresses for all the users for all the members
        managed by this roster.
        """)

    def get_member(email):
        """Get the member for the given address.

        *Note* that it is possible for an email to be subscribed to a
        mailing list twice, once through its explicit address and once
        indirectly through a user's preferred address.  In this case,
        this API always returns the explicit address.  Use
        ``get_memberships()`` to return them all.
        Likewise, an email can be a member of the administrators roster,
        both as an owner and as a moderator.  In this case, this API will
        return the owner.

        :param email: The email address to search for.
        :type email: string
        :return: The member if found, otherwise None
        :rtype: `IMember` or None
        """

    def get_memberships(email):
        """Get the memberships for the given address.

        :param email: The email address to search for.
        :type email: string
        :return: All the memberships associated with this email address.
        :rtype: sequence of length 0, 1, or 2 of ``IMember``
        """



# Copyright (C) 2007-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Implementations of the pending requests interfaces."""

from lazr.config import as_timedelta
from mailman.config import config
from mailman.database.model import Model
from mailman.database.transaction import dbconnection
from mailman.database.types import Enum, SAUnicode
from mailman.interfaces.pending import IPendable, IPendings
from mailman.interfaces.requests import IListRequests, RequestType
from mailman.model.pending import Pended, PendedKeyValue
from mailman.utilities.queries import QuerySequence
from pickle import dumps, loads
from public import public
from sqlalchemy import Column, ForeignKey, Integer, select
from sqlalchemy.orm import relationship
from zope.component import getUtility
from zope.interface import implementer


@public
@implementer(IPendable)
class DataPendable(dict):
    """See `IPendable`."""

    PEND_TYPE = 'data'

    def update(self, mapping):
        # Keys and values must be strings (unicodes, but bytes values are
        # accepted for now).  Any other types for keys are a programming
        # error.  If we find a non-SAUnicode value, pickle it and encode it in
        # such a way that it will be properly reconstituted when unpended.
        clean_mapping = {}
        for key, value in mapping.items():
            assert isinstance(key, (bytes, str))
            if not isinstance(value, str):
                key = '_pck_' + key
                value = dumps(value).decode('raw-unicode-escape')
            clean_mapping[key] = value
        super().update(clean_mapping)


@public
@implementer(IListRequests)
class ListRequests:
    """See `IListRequests`."""

    def __init__(self, mailing_list):
        self.mailing_list = mailing_list

    @property
    @dbconnection
    def count(self, store):
        return store.query(_Request).filter_by(
            mailing_list=self.mailing_list).count()

    @dbconnection
    def count_of(self, store, request_type):
        return store.query(_Request).filter_by(
            mailing_list=self.mailing_list, request_type=request_type).count()

    @property
    @dbconnection
    def held_requests(self, store):
        results = store.query(_Request).filter_by(
            mailing_list=self.mailing_list)
        yield from results

    @dbconnection
    def of_type(self, store, request_type):
        return (
            QuerySequence(
                store,
                select(_Request)
                .filter_by(
                    mailing_list=self.mailing_list, request_type=request_type)
                .order_by(_Request.id),
                )
            )

    @dbconnection
    def hold_request(self, store, request_type, key, data=None):
        # Check to make sure `request_type` is a valid Enum.
        if not isinstance(request_type, RequestType):
            try:
                request_type = RequestType(request_type)
            except ValueError:
                raise TypeError(request_type)

        if data is None:
            data_hash = None
        else:
            pendable = DataPendable()
            pendable.update(data)
            # MAS Use the moderator default lifetime.
            lifetime = as_timedelta(config.mailman.moderator_request_life)
            token = getUtility(IPendings).add(pendable, lifetime=lifetime)
            data_hash = token
        request = _Request(key, request_type, self.mailing_list, data_hash)
        store.add(request)
        # XXX The caller needs a valid id immediately, so flush the changes
        # now to the SA transaction context.  Otherwise .id would not be
        # valid.  Hopefully this has no unintended side-effects.
        store.flush()
        return request.id

    @dbconnection
    def get_request(self, store, request_id, request_type=None):
        result = store.query(_Request).get(request_id)
        if result is None or result.mailing_list != self.mailing_list:
            return None
        if request_type is not None and result.request_type != request_type:
            return None
        if result.data_hash is None:
            return result.key, None
        pendable = getUtility(IPendings).confirm(
            result.data_hash, expunge=False)
        if pendable is None:
            return None
        data = dict()
        # Unpickle any non-SAUnicode values.
        for key, value in pendable.items():
            if key.startswith('_pck_'):
                data[key[5:]] = loads(value.encode('raw-unicode-escape'))
            else:
                data[key] = value
        # Some APIs need the request type.
        data['_request_type'] = result.request_type.name
        return result.key, data

    @dbconnection
    def delete_request(self, store, request_id):
        request = store.query(_Request).get(request_id)
        if request is None:
            raise KeyError(request_id)
        # Throw away the pended data.
        getUtility(IPendings).confirm(request.data_hash)
        store.delete(request)

    @dbconnection
    def clear(self, store):
        for token, pendable in getUtility(IPendings).find(
                mlist=self.mailing_list,
                confirm=False):
            pended = store.query(Pended).filter_by(token=token).first()
            store.query(PendedKeyValue).filter_by(pended_id=pended.id).delete()
            store.delete(pended)


class _Request(Model):
    """Table for mailing list hold requests."""

    __tablename__ = '_request'

    id = Column(Integer, primary_key=True)
    key = Column(SAUnicode)
    request_type = Column(Enum(RequestType))
    data_hash = Column(SAUnicode)

    mailing_list_id = Column(Integer, ForeignKey('mailinglist.id'), index=True)
    mailing_list = relationship('MailingList')

    def __init__(self, key, request_type, mailing_list, data_hash):
        super().__init__()
        self.key = key
        self.request_type = request_type
        self.mailing_list = mailing_list
        self.data_hash = data_hash



# Copyright (C) 2007-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Interface for email address related information."""

from mailman.interfaces.errors import MailmanError
from public import public
from zope.interface import Attribute, Interface


@public
class EmailError(MailmanError):
    """A generic text email address-related error occurred."""

    def __init__(self, email):
        super().__init__()
        self.email = email

    def __str__(self):
        return self.email


@public
class AddressError(MailmanError):
    """A generic IAddress-related error occurred."""

    def __init__(self, address):
        super().__init__()
        self.address = address

    def __str__(self):
        return str(self.address)


@public
class ExistingAddressError(AddressError):
    """The given email address already exists."""


@public
class AddressAlreadyLinkedError(AddressError):
    """The address is already linked to a user."""


@public
class AddressNotLinkedError(AddressError):
    """The address is not linked to the user."""


@public
class InvalidEmailAddressError(EmailError):
    """Email address is invalid."""


@public
class AddressVerificationEvent:
    """Triggered when an address gets verified or unverified."""

    def __init__(self, address):
        self.address = address

    def __str__(self):
        return '<{} {} {}>'.format(
            self.__class__.__name__,
            self.address.email,
            ('unverified' if self.address.verified_on is None
             else self.address.verified_on))


@public
class IAddress(Interface):
    """Email address related information."""

    email = Attribute(
        """Read-only text email address.""")

    original_email = Attribute(
        """Read-only original case-preserved email address.

        For almost all intents and purposes, email addresses in Mailman are
        case insensitive, however because RFC 2821 allows for case sensitive
        local parts, Mailman preserves the case of the original email address
        when delivering a message to the user.

        `original_email` will be the same as `email` if the original email
        address was all lower case.  Otherwise `original_email` will be the
        case preserved email address; `email` will always be lower case.
        """)

    display_name = Attribute(
        """Optional display name associated with the email address.""")

    registered_on = Attribute(
        """The date and time at which this email address was registered.

        Registeration is really the date at which this address became known to
        us.  It may have been explicitly registered by a user, or it may have
        been implicitly registered, e.g. by showing up in a nonmember
        posting.""")

    verified_on = Attribute(
        """The date and time at which this email address was validated, or
        None if the email address has not yet been validated.  The specific
        method of validation is not defined here.""")

    preferences = Attribute(
        """This address's preferences.""")


@public
class IEmailValidator(Interface):
    """An email validator."""

    def is_valid(email):
        """Check if an email address if valid.

        :param email: A text email address.
        :type email: str
        :return: A flag indicating whether the email address is okay or not.
        :rtype: bool
        """

    def validate(email):
        """Validate an email address.

        :param email: A text email address.
        :type email: str
        :raise InvalidEmailAddressError: when `email` is deemed invalid.
        """



# Copyright (C) 2007-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

import re
import sys

from setuptools import find_packages, setup
from string import Template


if sys.hexversion < 0x30600f0:
    print('Mailman requires at least Python 3.6')
    sys.exit(1)


# Calculate the version number without importing the mailman package.
with open('src/mailman/version.py') as fp:
    for line in fp:
        mo = re.match("VERSION = '(?P<version>[^']+?)'", line)
        if mo:
            __version__ = mo.group('version')
            break
    else:
        print('No version number found')
        sys.exit(1)



# Ensure that all the .mo files are generated from the corresponding .po file.
# This procedure needs to be made sane, probably when the language packs are
# properly split out.

# Create the .mo files from the .po files.  There may be errors and warnings
# here and that could cause the digester.txt test to fail.
## start_dir = os.path.dirname('src/mailman/messages')
## for dirpath, dirnames, filenames in os.walk(start_dir):
##     for filename in filenames:
##         po_file = os.path.join(dirpath, filename)
##         basename, ext = os.path.splitext(po_file)
##         if ext <> '.po':
##             continue
##         mo_file = basename + '.mo'
##         if (not os.path.exists(mo_file) or
##             os.path.getmtime(po_file) > os.path.getmtime(mo_file)):
##             # The mo file doesn't exist or is older than the po file.
##             os.system('msgfmt -o %s %s' % (mo_file, po_file))



# XXX The 'bin/' prefix here should be configurable.
template = Template('$script = mailman.bin.$script:main')
scripts = set(
    template.substitute(script=script)
    for script in ('mailman', 'runner', 'master')
    )



setup(
    name            = 'mailman',
    version         = __version__,
    description     = 'Mailman -- the GNU mailing list manager',
    long_description= """\
This is GNU Mailman, a mailing list management system distributed under the
terms of the GNU General Public License (GPL) version 3 or later.  The name of
this software is spelled 'Mailman' with a leading capital 'M' but with a lower
case second 'm'.  Any other spelling is incorrect.""",
    author          = 'The Mailman Developers',
    author_email    = 'mailman-developers@python.org',
    license         = 'GPLv3',
    url             = 'https://www.list.org',
    project_urls={
        'Documentation': 'https://docs.mailman3.org/projects/mailman/en/latest/README.html',
        'Source': 'https://gitlab.com/mailman/mailman.git',
        'Tracker': 'https://gitlab.com/mailman/mailman/-/issues',
        },
    keywords        = 'email',
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: '
        'GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Communications :: Email :: Mailing List Servers',
        'Topic :: Communications :: Usenet News',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
        ],
    packages        = find_packages('src'),
    package_dir     = {'': 'src'},
    include_package_data = True,
    entry_points    = {
        'console_scripts' : list(scripts),
        },
    install_requires = [
        'aiosmtpd>=1.4.3',
        'alembic>=1.6.2,!=1.7.0',
        'atpublic',
        'authheaders>=0.14.0',
        'authres>=1.0.1',
        'click>=8.0.0',
        'dnspython>=1.14.0',
        'falcon>=3.0.0',
        'flufl.bounce>=4.0',
        'flufl.i18n>=3.2',
        'flufl.lock>=5.1',
        'importlib_resources>=1.1.0',
        'gunicorn',
        'lazr.config',
        'python-dateutil>=2.0',
        'passlib',
        'requests',
        'sqlalchemy>=1.4.0',
        'zope.component',
        'zope.configuration',
        'zope.event',
        'zope.interface>=5.0',
        ],
    )

# flake8: noqa



# Copyright (C) 2015-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""API version tests."""

import unittest

from mailman.core.system import system
from mailman.testing.helpers import call_api
from mailman.testing.layers import RESTLayer
from urllib.error import HTTPError


class TestAPIVersion(unittest.TestCase):
    layer = RESTLayer

    def test_api_31(self):
        # API version 3.1 was introduced in Mailman 3.1.
        url = 'http://localhost:9001/3.1/system'
        new = '{}/versions'.format(url)
        json, response = call_api(url)
        self.assertEqual(json['mailman_version'], system.mailman_version)
        self.assertEqual(json['python_version'], system.python_version)
        self.assertEqual(json['api_version'], '3.1')
        self.assertEqual(json['self_link'], new)

    def test_api_30(self):
        # API version 3.0 is still supported.
        url = 'http://localhost:9001/3.0/system'
        new = '{}/versions'.format(url)
        json, response = call_api(url)
        self.assertEqual(json['mailman_version'], system.mailman_version)
        self.assertEqual(json['python_version'], system.python_version)
        self.assertEqual(json['api_version'], '3.0')
        self.assertEqual(json['self_link'], new)

    def test_bad_api(self):
        # There is no API version earlier than 3.0.
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/2.9/system')
        self.assertEqual(cm.exception.code, 404)



# Copyright (C) 2014-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test the to_digest handler."""

import os
import unittest

from mailman.app.lifecycle import create_list
from mailman.handlers.to_digest import ToDigest
from mailman.testing.helpers import specialized_message_from_string as mfs
from mailman.testing.layers import ConfigLayer


class TestToDigest(unittest.TestCase):
    """Test the to_digest handler."""

    layer = ConfigLayer

    def setUp(self):
        self._mlist = create_list('test@example.com')
        self._msg = mfs("""\
From: anne@example.com
To: test@example.com
Subject: A disposable message
Message-ID: <ant>

""")
        self._handler = ToDigest()

    def test_unicode_message(self):
        # LP: #1170347 - The message has non-ascii characters in its payload,
        # but no charset (encoding) is defined e.g. in a Content-Type header.
        self._msg.set_payload(b'non-ascii chars \xc3\xa9 \xc3\xa8 \xc3\xa7')
        self._msg['X-Test'] = 'dummy'
        self._handler.process(self._mlist, self._msg, {})
        # Make sure the digest mbox is not empty.
        mailbox_path = os.path.join(self._mlist.data_path, 'digest.mmdf')
        self.assertGreater(os.path.getsize(mailbox_path), 0)

    def test_absent_parent_directory(self):
        # Remove parent directory to setup for testing
        os.rmdir(self._mlist.data_path)
        # Call process function
        self._msg.set_payload('test')
        self._msg['X-Test'] = 'dummy'
        self._handler.process(self._mlist, self._msg, {})
        # Assert digest.mmdf parent directory is present
        self.assertTrue(os.path.exists(self._mlist.data_path))



# Copyright (C) 2009-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""The 'addmembers' subcommand."""

import sys
import click

from email.utils import formataddr, parseaddr
from mailman.core.i18n import _
from mailman.database.transaction import transactional
from mailman.interfaces.address import IEmailValidator
from mailman.interfaces.command import ICLISubCommand
from mailman.interfaces.listmanager import IListManager
from mailman.interfaces.member import (
    AlreadySubscribedError,
    DeliveryMode,
    DeliveryStatus,
    MembershipIsBannedError,
)
from mailman.interfaces.subscriptions import (
    ISubscriptionManager,
    SubscriptionPendingError,
)
from mailman.interfaces.usermanager import IUserManager
from mailman.utilities.options import I18nCommand
from public import public
from zope.component import getUtility
from zope.interface import implementer


def get_addr(display_name, email, user_manager):
    """Return an existing address record if available, otherwise make one."""
    addr = user_manager.get_address(email)
    if addr is not None:
        # We have an address with this email.  Return that.
        return addr
    # Unknown email.  Create an address for this.
    # XXX Should we be making a user instead?
    return user_manager.create_address(email, display_name)


@transactional
def add_members(mlist, in_fp, delivery, invite, welcome_msg):
    """Add members to a mailing list."""
    user_manager = getUtility(IUserManager)
    registrar = ISubscriptionManager(mlist)
    email_validator = getUtility(IEmailValidator)
    for line in in_fp:
        # Ignore blank lines and lines that start with a '#'.
        if line.startswith('#') or len(line.strip()) == 0:
            continue
        # Parse the line and ensure that the values are unicodes.
        display_name, email = parseaddr(line)
        # parseaddr can return invalid emails.  E.g. parseaddr('foobar@')
        # returns ('', 'foobar@') in python 3.6.7 and 3.7.1 so check validity.
        if not email_validator.is_valid(email):
            line = line.strip()
            print(_('Cannot parse as valid email address (skipping): ${line}'),
                  file=sys.stderr)
            continue
        subscriber = get_addr(display_name, email, user_manager)
        # For error messages.
        email = formataddr((display_name, email))
        delivery_status = DeliveryStatus.enabled
        if delivery is None or delivery == 'regular' or delivery == 'disabled':
            delivery_mode = DeliveryMode.regular
            if delivery == 'disabled':
                delivery_status = DeliveryStatus.by_moderator
        elif delivery == 'mime':
            delivery_mode = DeliveryMode.mime_digests
        elif delivery == 'plain':
            delivery_mode = DeliveryMode.plaintext_digests
        elif delivery == 'summary':
            delivery_mode = DeliveryMode.summary_digests
        try:
            member = registrar.register(
                subscriber,
                pre_verified=True,
                pre_approved=True,
                pre_confirmed=True,
                invitation=invite,
                send_welcome_message=welcome_msg)[2]
            if member is not None:
                member.preferences.delivery_status = delivery_status
                member.preferences.delivery_mode = delivery_mode
        except AlreadySubscribedError:
            # It's okay if the address is already subscribed, just print a
            # warning and continue.
            print(_('Already subscribed (skipping): ${email}'),
                  file=sys.stderr)
        except MembershipIsBannedError:
            print(_('Membership is banned (skipping): ${email}'),
                  file=sys.stderr)
        except SubscriptionPendingError:
            print(_('Subscription already pending (skipping): ${email}'),
                  file=sys.stderr)


@click.command(
    cls=I18nCommand,
    help=_("""\
    Add all member addresses in FILENAME with delivery mode as specified
    with -d/--delivery.  FILENAME can be '-' to indicate standard input.
    Blank lines and lines that start with a '#' are ignored.
    """))
@click.option(
    '--delivery', '-d',
    type=click.Choice(('regular', 'mime', 'plain', 'summary', 'disabled')),
    help=_("""\
    Set the added members delivery mode to 'regular', 'mime', 'plain',
    'summary' or 'disabled'.  I.e., one of regular, three modes of digest
    or no delivery.  If not given, the default is regular.  Ignored for invited
    members."""))
@click.option(
    '--invite', '-i',
    is_flag=True, default=False,
    help=_("""\
    Send the added members an invitation rather than immediately adding them.
    """))
@click.option(
    '--welcome-msg/--no-welcome-msg', '-w/-W', 'welcome_msg', default=None,
    help=_("""\
    Override the list's setting for send_welcome_message."""))
@click.argument('in_fp', metavar='FILENAME', type=click.File(encoding='utf-8'))
@click.argument('listspec')
@click.pass_context
def addmembers(ctx, in_fp, delivery, invite, welcome_msg, listspec):
    """Add members to a mailing list."""
    mlist = getUtility(IListManager).get(listspec)
    if mlist is None:
        ctx.fail(_('No such list: ${listspec}'))
    add_members(mlist, in_fp, delivery, invite, welcome_msg)


@public
@implementer(ICLISubCommand)
class AddMembers:
    name = 'addmembers'
    command = addmembers



# Copyright (C) 2015-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test the `mailman members` command."""

import unittest

from click.testing import CliRunner
from mailman.commands.cli_changeaddress import changeaddress
from mailman.interfaces.usermanager import IUserManager
from mailman.testing.layers import ConfigLayer
from zope.component import getUtility


class TestCLIFindMember(unittest.TestCase):
    layer = ConfigLayer

    def setUp(self):
        self._um = getUtility(IUserManager)
        self._adr1 = self._um.create_address('anne@example.com', 'Anne Person')
        self._adr2 = self._um.create_address('bart@example.com', 'Bart Person')
        self._command = CliRunner()

    def test_basic_change(self):
        # Test a simple change of address.
        result = self._command.invoke(changeaddress, ('anne@example.com',
                                                      'anne@example.net'))
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.output, """\
Address changed from anne@example.com to anne@example.net.
""")
        self.assertIsNone(self._um.get_address('anne@example.com'))
        self.assertIn('<Address: Anne Person <anne@example.net>',
                      repr(self._um.get_address('anne@example.net')))

    def test_change_to_existing_fails(self):
        # Test trying to change to and existing address.
        result = self._command.invoke(changeaddress, ('anne@example.com',
                                                      'bart@example.com'))
        self.assertEqual(result.exit_code, 2)
        self.assertEqual(result.output, """\
Usage: changeaddress [OPTIONS] OLD_ADDRESS NEW_ADDRESS
Try 'changeaddress --help' for help.

Error: Address bart@example.com already exists; can't change.
""")

    def test_change_from_nonexisting_fails(self):
        # Test trying to change a nonexistant address.
        result = self._command.invoke(changeaddress, ('anne@example.net',
                                                      'anne@example.edu'))
        self.assertEqual(result.exit_code, 2)
        self.assertEqual(result.output, """\
Usage: changeaddress [OPTIONS] OLD_ADDRESS NEW_ADDRESS
Try 'changeaddress --help' for help.

Error: Address anne@example.net not found.
""")

    def test_case_change_only(self):
        # Test that we can change case of existing address.
        result = self._command.invoke(changeaddress, ('anne@example.com',
                                                      'Anne@example.com'))
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.output, """\
Address changed from anne@example.com to Anne@example.com.
""")
        self.assertIn('<Address: Anne Person <Anne@example.com>',
                      repr(self._um.get_address('anne@example.com')))

    def test_change_to_invalid_address(self):
        # Test trying to change to an invalid address.
        result = self._command.invoke(changeaddress, ('anne@example.com',
                                                      'invalid_address'))
        self.assertEqual(result.exit_code, 2)
        self.assertEqual(result.output, """\
Usage: changeaddress [OPTIONS] OLD_ADDRESS NEW_ADDRESS
Try 'changeaddress --help' for help.

Error: Address invalid_address is not a valid email address.
""")

    def test_change_not_different(self):
        # Test no change.
        result = self._command.invoke(changeaddress, ('anne@example.com',
                                                      'anne@example.com'))
        self.assertEqual(result.exit_code, 2)
        self.assertEqual(result.output, """\
Usage: changeaddress [OPTIONS] OLD_ADDRESS NEW_ADDRESS
Try 'changeaddress --help' for help.

Error: Addresses are not different.  Nothing to change.
""")



# Copyright (C) 2015-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test mailman command utilities."""

import unittest

from click.testing import CliRunner
from datetime import timedelta
from importlib_resources import path
from mailman.app.lifecycle import create_list
from mailman.bin.mailman import main
from mailman.config import config
from mailman.database.transaction import transaction
from mailman.interfaces.command import ICLISubCommand
from mailman.testing.layers import ConfigLayer
from mailman.utilities.datetime import now
from mailman.utilities.modules import add_components
from unittest.mock import patch


def mock_euid():
    return 0


class TestMailmanCommand(unittest.TestCase):
    layer = ConfigLayer

    def setUp(self):
        self._command = CliRunner()

    def test_mailman_command_config(self):
        with path('mailman.testing', 'testing.cfg') as config_path:
            with patch('mailman.bin.mailman.initialize') as init:
                self._command.invoke(main, ('-C', str(config_path), 'info'))
            init.assert_called_once_with(str(config_path))

    def test_mailman_command_no_config(self):
        with patch('mailman.bin.mailman.initialize') as init:
            self._command.invoke(main, ('info',))
        init.assert_called_once_with(None)

    @patch('mailman.bin.mailman.os')
    @patch('mailman.bin.mailman.initialize')
    def test_mailman_command_without_subcommand_prints_help(
            self, mock_os, mock):
        mock_os.geteuid.return_value = 1000
        # Issue #137: Running `mailman` without a subcommand raises an
        # AttributeError.
        result = self._command.invoke(main, [])
        lines = result.output.splitlines()
        # "main" instead of "mailman" because of the way the click runner
        # works.  It does actually show the correct program when run from the
        # command line.
        self.assertEqual(lines[0], 'Usage: main [OPTIONS] COMMAND [ARGS]...')
        # The help output includes a list of subcommands, in sorted order.
        commands = {}
        add_components('commands', ICLISubCommand, commands)
        help_commands = list(
            line.split()[0].strip()
            for line in lines[-len(commands):]
            )
        self.assertEqual(sorted(commands), help_commands)

    @patch('mailman.bin.mailman.initialize')
    def test_mailman_command_with_bad_subcommand_prints_help(self, mock):
        # Issue #137: Running `mailman` without a subcommand raises an
        # AttributeError.
        result = self._command.invoke(main, ('not-a-subcommand',))
        lines = result.output.splitlines()
        # "main" instead of "mailman" because of the way the click runner
        # works.  It does actually show the correct program when run from the
        # command line.
        self.assertEqual(lines[0], 'Usage: main [OPTIONS] COMMAND [ARGS]...')

    @patch('mailman.bin.mailman.os')
    @patch('mailman.bin.mailman.initialize')
    def test_transaction_commit_after_successful_subcommand(
            self, mock_os, mock):
        mock_os.geteuid.return_value = 0
        # Issue #223: Subcommands which change the database need to commit or
        # abort the transaction.
        with transaction():
            mlist = create_list('ant@example.com')
            mlist.volume = 5
            mlist.next_digest_number = 3
            mlist.digest_last_sent_at = now() - timedelta(days=60)
        self._command.invoke(main, ('digests', '-b', '-l', 'ant@example.com'))
        # Clear the current transaction to force a database reload.
        config.db.abort()
        self.assertEqual(mlist.volume, 6)
        self.assertEqual(mlist.next_digest_number, 1)

    @patch('mailman.bin.mailman.initialize')
    @patch('mailman.commands.cli_digests.maybe_send_digest_now',
           side_effect=RuntimeError)
    def test_transaction_abort_after_failing_subcommand(self, mock1, mock2):
        with transaction():
            mlist = create_list('ant@example.com')
            mlist.volume = 5
            mlist.next_digest_number = 3
            mlist.digest_last_sent_at = now() - timedelta(days=60)
        self._command.invoke(
            main, ('digests', '-b', '-l', 'ant@example.com', '--send'))
        # Clear the current transaction to force a database reload.
        config.db.abort()
        # The volume and number haven't changed.
        self.assertEqual(mlist.volume, 5)
        self.assertEqual(mlist.next_digest_number, 3)

    @patch('mailman.bin.mailman.initialize')
    @patch('os.geteuid', mock_euid)
    def test_wont_run_as_root(self, mock):
        result = self._command.invoke(main)
        self.assertIn(
            'Error: If you are sure you want to run as root, '
            'specify --run-as-root.',
            result.output)
        self.assertNotEqual(result.exit_code, 0)

    @patch('mailman.bin.mailman.initialize')
    @patch('os.geteuid', mock_euid)
    def test_will_run_as_root_with_option(self, mock):
        result = self._command.invoke(main, ('--run-as-root'))
        self.assertNotIn('Error:', result.output)
        self.assertEqual(result.exit_code, 0)



# Copyright (C) 2006-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Model for addresses."""

from email.utils import formataddr
from mailman.database.model import Model
from mailman.database.types import SAUnicode, SAUnicode4Byte
from mailman.interfaces.address import (
    AddressVerificationEvent,
    IAddress,
    IEmailValidator,
)
from mailman.utilities.datetime import now
from public import public
from sqlalchemy import Column, DateTime, ForeignKey, Integer
from sqlalchemy.orm import backref, relationship
from zope.component import getUtility
from zope.event import notify
from zope.interface import implementer


@public
@implementer(IAddress)
class Address(Model):
    """See `IAddress`."""

    __tablename__ = 'address'

    id = Column(Integer, primary_key=True)
    email = Column(SAUnicode, index=True, unique=True)
    _original = Column(SAUnicode)
    display_name = Column(SAUnicode4Byte)
    _verified_on = Column('verified_on', DateTime)
    registered_on = Column(DateTime)

    user_id = Column(Integer, ForeignKey('user.id'), index=True)

    preferences_id = Column(Integer, ForeignKey('preferences.id'), index=True)
    preferences = relationship(
        'Preferences', backref=backref('address', uselist=False))

    def __init__(self, email, display_name):
        super().__init__()
        getUtility(IEmailValidator).validate(email)
        lower_case = email.lower()
        self.email = lower_case
        self.display_name = display_name
        self._original = (None if lower_case == email else email)
        self.registered_on = now()

    def __str__(self):
        addr = (self.email if self._original is None else self._original)
        return formataddr((self.display_name, addr))

    def __repr__(self):
        verified = ('verified' if self.verified_on else 'not verified')
        address_str = str(self)
        if self._original is None:
            return '<Address: {} [{}] at {:#x}>'.format(
                address_str, verified, id(self))
        else:
            return '<Address: {} [{}] key: {} at {:#x}>'.format(
                address_str, verified, self.email, id(self))

    @property
    def verified_on(self):
        return self._verified_on

    @verified_on.setter
    def verified_on(self, timestamp):
        self._verified_on = timestamp
        notify(AddressVerificationEvent(self))

    @property
    def original_email(self):
        return (self.email if self._original is None else self._original)



# Copyright (C) 2015-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test the rfc_2369 handler."""

import unittest

from mailman.app.lifecycle import create_list
from mailman.config import config
from mailman.handlers import rfc_2369
from mailman.interfaces.archiver import ArchivePolicy, IArchiver
from mailman.testing.helpers import (
    LogFileMark,
    specialized_message_from_string as mfs,
)
from mailman.testing.layers import ConfigLayer
from urllib.parse import urljoin
from zope.interface import implementer


@implementer(IArchiver)
class DummyArchiver:
    """An example archiver which does nothing but return URLs."""
    name = 'dummy'

    def list_url(self, mlist):
        """See `IArchiver`."""
        return 'http://{}'.format(mlist.mail_host)

    def permalink(self, mlist, msg):
        """See `IArchiver`."""
        message_id_hash = msg.get('message-id-hash')
        if message_id_hash is None:
            return None
        return urljoin(self.list_url(mlist), message_id_hash)

    @staticmethod
    def archive_message(mlist, message):
        return None


@implementer(IArchiver)
class BrokenArchiver:
    """An archiver that has some broken methods."""

    name = 'broken'

    def list_url(self, mlist):
        raise RuntimeError('Cannot get list URL')

    def permalink(self, mlist, msg):
        raise RuntimeError('Cannot get permalink')

    @staticmethod
    def archive_message(mlist, message):
        raise RuntimeError('Cannot archive message')


class TestRFC2369(unittest.TestCase):
    """Test the rfc_2369 handler."""

    layer = ConfigLayer

    def setUp(self):
        config.push('no_archivers', """
        [archiver.prototype]
        enable: no
        [archiver.mail_archive]
        enable: no
        [archiver.mhonarc]
        enable: no
        [archiver.pipermail]
        enable: no
        """)
        self.addCleanup(config.pop, 'no_archivers')
        self._mlist = create_list('test@example.com')
        self._mlist.archive_policy = ArchivePolicy.public
        self._msg = mfs("""\
From: aperson@example.com
Message-ID: <first>
Message-ID-Hash: 4CMWUN6BHVCMHMDAOSJZ2Q72G5M32MWB

Dummy text
""")

    def test_add_headers(self):
        # Test the addition of the Archived-At and List-Archive headers.
        config.push('archiver', """
        [archiver.dummy]
        class: {}.DummyArchiver
        enable: yes
        """.format(DummyArchiver.__module__))
        self.addCleanup(config.pop, 'archiver')
        rfc_2369.process(self._mlist, self._msg, {})
        self.assertEqual(
            self._msg.get_all('List-Archive'), ['<http://example.com>'])
        self.assertEqual(
            self._msg.get_all('Archived-At'),
            ['<http://example.com/4CMWUN6BHVCMHMDAOSJZ2Q72G5M32MWB>'])

    def test_prototype_no_url(self):
        # The prototype archiver is not web-based, it must not return URLs
        config.push('archiver', """
        [archiver.prototype]
        enable: yes
        """)
        self.addCleanup(config.pop, 'archiver')
        rfc_2369.process(self._mlist, self._msg, {})
        self.assertNotIn('Archived-At', self._msg)
        self.assertNotIn('List-Archive', self._msg)

    def test_not_archived(self):
        # Messages sent to non-archived lists must not get the added headers.
        self._mlist.archive_policy = ArchivePolicy.never
        config.push('archiver', """
        [archiver.dummy]
        class: {}.DummyArchiver
        enable: yes
        """.format(DummyArchiver.__module__))
        self.addCleanup(config.pop, 'archiver')
        rfc_2369.process(self._mlist, self._msg, {})
        self.assertNotIn('List-Archive', self._msg)
        self.assertNotIn('Archived-At', self._msg)

    def test_broken_archiver(self):
        # GL issue #208 - IArchive messages raise exceptions, breaking the
        # rfc-2369 handler and shunting messages.
        config.push('archiver', """
        [archiver.broken]
        class: {}.BrokenArchiver
        enable: yes
        """.format(BrokenArchiver.__module__))
        self.addCleanup(config.pop, 'archiver')
        mark = LogFileMark('mailman.archiver')
        rfc_2369.process(self._mlist, self._msg, {})
        log_messages = mark.read()
        # Because .list_url() was broken, there will be no List-Archive header.
        self.assertIsNone(self._msg.get('list-archive'))
        self.assertIn('Exception in "broken" archiver', log_messages)
        self.assertIn('RuntimeError: Cannot get list URL', log_messages)
        # Because .permalink() was broken, there will be no Archived-At header.
        self.assertIsNone(self._msg.get('archived-at'))
        self.assertIn('Exception in "broken" archiver', log_messages)
        self.assertIn('RuntimeError: Cannot get permalink', log_messages)

    def test_add_other_headers(self):
        # Test that other RFC-2369 headers are added.
        self._mlist.include_rfc2369_headers = True
        rfc_2369.process(self._mlist, self._msg, {})
        self.assertEqual(
            self._msg.get_all('List-Help'),
            ['<mailto:{}?subject=help>'.format(self._mlist.request_address)])
        self.assertEqual(
            self._msg.get_all('List-Owner'),
            ['<mailto:{}>'.format(self._mlist.owner_address)])
        self.assertEqual(
            self._msg.get_all('List-Unsubscribe'),
            ['<mailto:{}>'.format(self._mlist.leave_address)])
        self.assertEqual(
            self._msg.get_all('List-Subscribe'),
            ['<mailto:{}>'.format(self._mlist.join_address)])

    def test_dont_add_other_headers(self):
        # Test that other RFC-2369 headers are not added if not requested.
        self._mlist.include_rfc2369_headers = False
        rfc_2369.process(self._mlist, self._msg, {})
        self.assertIsNone(self._msg.get_all('List-Help'))
        self.assertIsNone(self._msg.get_all('List-Owner'))
        self.assertIsNone(self._msg.get_all('List-Unsubscribe'))
        self.assertIsNone(self._msg.get_all('List-Subscribe'))



# Copyright (C) 2006-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""A virgin pipeline."""

from mailman.core.i18n import _
from mailman.pipelines.base import BasePipeline
from public import public


@public
class VirginPipeline(BasePipeline):
    """The processing pipeline for virgin messages.

    Virgin messages are those that are crafted internally by Mailman.
    """
    name = 'virgin'
    description = _('The virgin queue pipeline.')

    _default_handlers = (
        'cook-headers',
        'rfc-2369',
        'to-outgoing',
        )
